from typing import Generic, TypeVar, Type, Union, List

from hatsudenki.packages.field import BaseHatsudenkiField, T

K = TypeVar('K')


class ReferenceMasterOneField(Generic[K], BaseHatsudenkiField['ReferenceMasterOneField.Value']):
    TypeStr = 'S'

    class Value(Generic[K]):
        def __init__(self, name: str, to: Type[K], label: str, parent=None):
            from hatsudenki.packages.table.solo import SoloHatsudenkiTable
            self.parent: SoloHatsudenkiTable = parent
            self.label = label
            self.name = name
            self.to = to

        def is_empty(self):
            return self.label is None or len(self.label) is 0

        def to_string(self):
            return self.label

        def update_key(self):
            if self.parent is None:
                return
            setattr(self.parent, self.name, self)

        @property
        def value(self):
            return self.label

        def resolved_value(self) -> K:
            return self.to.get_by_cursor(self.label)

        def connect(self, t: K):
            if t is None:
                self.label = None
            else:
                self.label = t.one_cursor
            self.update_key()

        def __str__(self):
            return self.to_string()

    def __init__(self, to: Type[K], **kwargs):
        super().__init__(**kwargs)
        self.to = to

    @property
    def to_table(self):
        return self.to

    def get_data(self, val: Union[str, K, None], parent=None) -> Value[K]:
        tp = type(val)
        if tp is str:
            val = val
        elif tp is self.to_table:
            val = val.one_cursor
        elif val is None:
            val = None
        elif tp is self.__class__.Value:
            return val
        else:
            raise Exception(f'invalid keys.{val}')
        return self.__class__.Value(self.name, self.to_table, val, parent)

    def get_data_from_dict(self, v: dict, table=None) -> Value[K]:
        return super().get_data_from_dict(v, table)

    @classmethod
    def is_empty(cls, value):
        if value is None:
            return True
        return value.is_empty()

    def to_string(self, val: Union[str, K, Value]):
        if val is None:
            return None
        if isinstance(val, ReferenceMasterOneField[K].Value):
            return val.to_string()
        if type(val) is str:
            return val
        if isinstance(val, self.to):
            return val.one_cursor
        raise Exception(f'invalid key type {val}')

    def serialize(self, value: T, table=None):
        s = value.to_string()
        if s:
            return {'S': s}
        return None

    def deserialize(self, value, table=None):
        # s = super().deserialize(value, table)
        if value:
            return self.get_data(value['S'], table)
        return self.__class__.Value(self.name, self.to_table, None, table)


class ReferenceMasterManyField(ReferenceMasterOneField[K]):
    def __init__(self, to: Type[K], **kwargs):
        super().__init__(to=to, **kwargs)

    class Value(ReferenceMasterOneField.Value, Generic[K]):
        def resolved_value(self) -> List[K]:
            return self.to.find_by_cursor(self.label)

        def connect(self, t: K):
            self.label = t.many_cursor
            self.update_key()

    def get_data(self, val: Union[str, K, None], parent=None) -> Value[K]:
        if type(val) is str:
            val = val
        elif type(val) is self.to_table:
            val = val.many_cursor
        elif val is None:
            val = None
        else:
            raise Exception('invalid keys.')
        return self.__class__.Value(self.name, self.to_table, val, parent)

    def get_data_from_dict(self, v: dict, table=None) -> Value[K]:
        return super().get_data_from_dict(v, table)


class ReferenceDynamoOneField(ReferenceMasterOneField[K]):
    TypeStr = 'S'

    class Value(ReferenceMasterOneField.Value, Generic[K]):

        async def resolved_value(self) -> K:
            return await self.to.query_by_cursor(self.label)

        def clear(self):
            self.label = None
            self.update_key()

        def connect(self, t: K):
            if t is None:
                self.clear()
            else:
                super().connect(t)

    def __init__(self, to: str, **kwargs):
        super().__init__(to=None, **kwargs)
        self.to = to

    @property
    def to_table(self):
        from hatsudenki.packages.manager.table import TableManager
        return TableManager.get_by_table_name(self.to)

    def get_data(self, val: Union[str, K, None], parent=None) -> Value[K]:
        return super().get_data(val, parent)

    def get_data_from_dict(self, v: dict, table=None) -> Value[K]:
        return super().get_data_from_dict(v, table)


class ReferenceDynamoManyField(ReferenceDynamoOneField[K]):
    TypeStr = 'S'

    class Value(ReferenceMasterOneField.Value, Generic[K]):
        async def resolved_value(self) -> List[K]:
            return await self.to.query_list_by_cursor(self.label)

        def connect(self, t: K):
            self.label = t.many_cursor
            self.update_key()

    @property
    def to_table(self):
        from hatsudenki.packages.manager.table import TableManager
        return TableManager.get_by_table_name(self.to)

    def get_data(self, val: Union[str, K, None], parent=None) -> Value[K]:
        if type(val) is str:
            val = val
        elif type(val) is self.to_table:
            val = val.many_cursor
        elif val is None:
            val = None
        else:
            raise Exception('invalid keys.')
        return self.__class__.Value(self.name, self.to_table, val, parent)

    def get_data_from_dict(self, v: dict, table=None) -> Value[K]:
        return super().get_data_from_dict(v, table)



from typing import Type, Iterator, List

from hatsudenki.packages.cache.base.memory.dynamic.sync.solo import SyncDynamicCacheBaseTableSolo
from hatsudenki.packages.cache.base.memory.multi import CacheBaseTableMulti
from hatsudenki.packages.cache.base.memory.solo import T


class SyncDynamicCacheBaseTableMulti(CacheBaseTableMulti, SyncDynamicCacheBaseTableSolo):
    _mapped = None
    _sorted_table = None

    @classmethod
    def invalidate(cls):
        super().invalidate()
        cls._sorted_table = None

    @classmethod
    def prepare(cls):
        super().prepare()
        cls._sorted_table = {}

    @classmethod
    def get(cls: Type[T], hash_key, range_key=None) -> T:
        cls._setup()
        return cls._get(hash_key, range_key)

    @classmethod
    def find(cls: Type[T], hash_key) -> List[T]:
        """
        HASHキーのみを指定してリスト
        :param hash_key: ハッシュキー
        :return: レコードの配列
        """
        cls._setup()
        return cls._find(hash_key)

    @classmethod
    def find_left(cls: Type[T], hash_key, range_key=None) -> T:
        cls._setup()
        return cls._find_left(hash_key, range_key)

    @classmethod
    def find_right(cls: Type[T], hash_key, range_key=None) -> T:
        cls._setup()
        return cls._find_right(hash_key, range_key)

    @classmethod
    def find_lte(cls: Type[T], hash_key: str, range_key) -> List[T]:
        """
        RANGEキー条件指定検索 - 以下

        :param hash_key: ハッシュキー
        :param range_key: 対象値
        :return: レコードの配列
        """
        cls._setup()
        return cls._find_lte(hash_key, range_key)

    @classmethod
    def find_lt(cls: Type[T], hash_key: str, range_key) -> List[T]:
        """
        RANGEキー条件指定検索 - 未満

        :param hash_key: ハッシュキー
        :param range_key: 対象値
        :return: レコードの配列
        """
        cls._setup()
        return cls._find_lt(hash_key, range_key)

    @classmethod
    def find_gte(cls: Type[T], hash_key: str, range_key) -> List[T]:
        """
        RANGEキー条件指定検索 - 以上

        :param hash_key: ハッシュキー
        :param range_key: 対象値
        :return: レコードの配列
        """
        cls._setup()
        return cls._find_gte(hash_key, range_key)

    @classmethod
    def find_gt(cls: Type[T], hash_key: str, range_key) -> List[T]:
        """
        RANGEキー条件指定検索 - より小さい

        :param hash_key: ハッシュキー
        :param range_key: 対象値
        :return: レコードの配列
        """
        cls._setup()
        return cls._find_gt(hash_key, range_key)

    @classmethod
    def find_between(cls: Type[T], hash_key: str, left_range, right_range) -> List[T]:
        """
        RANGEキー条件指定検索 - 範囲

        :param hash_key: ハッシュキー
        :param left_range: 左辺値
        :param right_range: 右辺値
        :return: レコードの配列
        """
        cls._setup()
        return cls._find_between(hash_key, left_range, right_range)

    @classmethod
    def get_by_cursor(cls: Type[T], cursor: str) -> T:
        cls._setup()
        return cls._get_by_cursor(cursor)

    @classmethod
    def find_by_cursor(cls: Type[T], cursor: str) -> List[T]:
        """
        集合カーソルを指定してリストを取得

        :param cursor: str
        :return: レコードの配列
        """
        cls._setup()
        return cls._find_by_cursor(cursor)

    @classmethod
    def iter(cls: Type[T]) -> Iterator[T]:
        cls._setup()
        return cls._iter()



import os
import re
import sys
import traceback
from itertools import chain
from typing import List, Union, Tuple


def snake_to_camel(value: str):
    if value is None:
        return None
    # まず全て小文字にする
    ret = value.lower()

    words = ret.split('_')
    words = [w.capitalize() for w in words]
    return ''.join(words)


def parse_type_str(type_str: str):
    brace = type_str.find('(')
    brace_str = None
    if brace != -1:
        s = type_str[:brace]
        brace_str = type_str[brace + 1:-1]
    else:
        s = type_str
    return (s, brace_str)


def path_to_snake(path: str, separator=os.sep):
    path = erase_ext(path)
    a = list(path.split(separator))
    if len(a) >= 2 and a[-1] == a[-2]:
        a.pop()
    res = chain.from_iterable([s.split('_') for s in a])
    r = '_'.join(res)
    return r


def erase_ext(file_name: str):
    return ''.join(file_name.split('.')[0:-1])


class ConsoleColor:
    BLACK = '\033[30m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    PURPLE = '\033[35m'
    CYAN = '\033[36m'
    WHITE = '\033[37m'
    END = '\033[0m'
    BOLD = '\038[1m'
    UNDERLINE = '\033[4m'
    INVISIBLE = '\033[08m'
    REVERCE = '\033[07m'


def color_print(color: str, text: str, indent=0):
    prefix = ''
    if indent >= 1:
        prefix = '  ' * (indent - 1) + (' └')
    text = prefix + text
    try:
        print(color + text + ConsoleColor.END)
    except UnicodeEncodeError:
        sys.stdout.buffer.write((text + '\n').encode('utf-8'))


class ToolOutput(object):
    TOOL_COLOR_LEVEL = [
        ConsoleColor.BLUE,
        ConsoleColor.GREEN,
        ConsoleColor.PURPLE,
        ConsoleColor.YELLOW,
        ConsoleColor.CYAN,
    ]

    out_indent = 0
    stack = []
    is_setup = False

    @classmethod
    def setup(cls):
        if not cls.is_setup:
            cls.is_setup = True
            cls.out_indent = 0
            cls.stack = []

    @classmethod
    def indent(cls, text=None):
        if cls.is_setup is False:
            return
        if text:
            cls.print(text)
        cls.out_indent += 1

    @classmethod
    def outdent(cls, text=None):
        if cls.is_setup is False:
            return
        cls.out_indent -= 1
        if text:
            cls.print(text)

    @classmethod
    def clear(cls):
        if cls.is_setup is False:
            return
        cls.out_indent = 0

    @classmethod
    def anchor(cls, text=None):
        if cls.is_setup is False:
            return
        cls.stack.append(cls.out_indent)
        if text is not None:
            cls.print(text)

    @classmethod
    def out(cls, text):
        if cls.is_setup is False:
            return
        cls.print(text, False)

    @classmethod
    def pop(cls, text=None):
        if cls.is_setup is False:
            return
        if text is not None:
            cls.print(text)
        cls.out_indent = cls.stack.pop()

    @classmethod
    def get_color(cls):
        return cls.TOOL_COLOR_LEVEL[cls.out_indent % len(cls.TOOL_COLOR_LEVEL)]

    @classmethod
    def print(cls, text: str, with_indent=True):
        if cls.is_setup is False:
            return
        for idx, t in enumerate(text.split('\n')):
            if idx is 0:
                color_print(cls.get_color(), t, cls.out_indent)
            else:
                color_print(cls.get_color(), '  ' * cls.out_indent + t)

        if with_indent:
            cls.indent()

    @classmethod
    def print_error(cls, text: str, with_indent=True):
        sys.stdout.flush()
        if cls.is_setup is False:
            return

        t2 = '【☆ERROR☆】 ' + text
        for idx, t in enumerate(t2.split('\n')):
            if idx is 0:
                color_print(ConsoleColor.RED, t, cls.out_indent)
            else:
                color_print(ConsoleColor.RED, '  ' * cls.out_indent + t)

        if with_indent:
            cls.indent()

    @classmethod
    def out_error(cls, text: str):
        if cls.is_setup is False:
            return
        cls.print_error(text, False)

    @classmethod
    def print_exc(cls, exc: Exception):
        sys.stdout.flush()
        if cls.is_setup is False:
            return
        color_print(ConsoleColor.RED, str(exc))
        color_print(ConsoleColor.RED, traceback.format_exc())

    @classmethod
    def print_with_anchor(cls, text: str):
        cls.anchor()
        cls.print(text)

    @classmethod
    def print_with_pop(cls, text: str):
        cls.print(text)
        cls.pop()

    @classmethod
    def print_list(cls, text_list: List[str], with_indent=True):
        [cls.print(t, False) for t in text_list]
        if with_indent:
            cls.indent()

    @classmethod
    def block(cls, label: str):
        class _():
            def __enter__(self):
                cls.anchor()
                cls.print(label)

            def __exit__(self, exc_type, exc_val, exc_tb):
                cls.pop()

        return _()


def quote(s: str, quote_str='\''):
    return quote_str + s + quote_str


def indent_str_list(str_list: List[str], indent=1, indent_str='    '):
    return [(indent_str * indent) + s for s in str_list]


def capitalize_join(strs: List[str], sep: str = ''):
    return sep.join([st.capitalize() for st in strs])


def parse_bracket(type_str: str):
    brace = type_str.find('(')
    brace_str = None
    if brace != -1:
        s = type_str[:brace]
        brace_str = type_str[brace + 1:-1]
    else:
        s = type_str
    return s, brace_str


IndentChild = Union[str, 'IndentString']
IndentUnit = Tuple[int, IndentChild]


class IndentString(object):

    def __init__(self, header: IndentChild = None, indent_str='    '):
        self._children: List[IndentUnit] = []
        self._indent_str = indent_str
        self._indent_level = 0

        if header is not None:
            self.add(header)
            self.indent()

    @property
    def line_num(self):
        return len(self._children)

    @property
    def indent_str(self):
        return self._indent_level * self._indent_str

    def add(self, *child: Union['IndentString', str]):
        [self._children.append((self._indent_level, c)) for c in child if c is not None]

    def blank_line(self, num=1):
        [self._children.append((0, '')) for _ in range(num)]

    def indent(self, header: Union['IndentString', str] = None):
        if header is not None:
            self.add(header)
        self._indent_level += 1

    def outdent(self, footer: Union['IndentString', str] = None):
        self._indent_level -= 1
        if footer is not None:
            self.add(footer)

    def render(self, offset_indent=0, sep='\n'):
        def _henkan(c: IndentUnit):
            lv, u = c
            lv += offset_indent

            tp = type(u)
            if tp is str:
                return (self._indent_str * lv) + u
            elif isinstance(u, IndentString):
                return u.render(offset_indent=lv)

        return sep.join(map(_henkan, self._children))


class TAGText(IndentString):

    def __init__(self, header: IndentChild = None, indent_str='    '):
        super().__init__(header, indent_str)
        self._anchor = []

    def tag(self, tag: str):
        self.indent(f'<{tag}>')
        self._anchor.append((self._indent_level, tag))

    def end_tag(self):
        lv, tag = self._anchor.pop()
        self.outdent(f'</{tag}>')

    def tag_with_content(self, tag: str, content: str):
        self.add(f'<{tag}>{content}</{tag}>')


_underscore1 = re.compile(r'(.)([A-Z][a-z]+)')
_underscore2 = re.compile('([a-z0-9])([A-Z])')


def camel_to_snake(s: str):
    subbed = _underscore1.sub(r'\1_\2', s)
    return _underscore2.sub(r'\1_\2', subbed).lower()



from pathlib import Path

from hatsudenki.packages.command.loader.base import LoaderBase, T
from hatsudenki.packages.command.master.exporter.pack.yaml import MasterExportYaml
from hatsudenki.packages.command.master.loader import MasterTableLoader


class MasterExportYamlLoader(LoaderBase[MasterExportYaml]):
    def __init__(self, base_path: Path, master_loader: MasterTableLoader):
        super().__init__(base_path)
        self.ext = '.yml'
        self.master_loader = master_loader

    def _load(self, path: Path) -> T:
        table = self.master_loader.get_by_table_name(path.name.replace('.yml', ''))
        return MasterExportYaml(self.base_path, path, table, self)

    def get_by_table_name(self, table_name: str):
        return self.get_from_path(self.base_path / (table_name + '.yml'))



from datetime import datetime
from uuid import UUID


def primal_deserializer(d):
    """
    基本型のデシリアライザ。DynamoDBからPythonに変換する。

    :param d: 対象値
    :return: Pythonで表現された値
    """

    k = list(d.keys())[0]
    v = d[k]
    if k == 'N':
        return int(v)
    if k in {'SS', 'BS'}:
        return {vv for vv in v}
    if k == 'NS':
        return {int(i) for i in v}
    if k == 'M':
        return {k: primal_deserializer(vv) for k, vv in v.items()}
    if k == 'L':
        return [primal_deserializer(vv) for vv in v]
    if k in {'B', 'BOOL'}:
        return v

    return d[k]


def _to_type_string(val):
    """
    型からDynamoDBシグネチャに変換する。

    :param val: 対象値
    :return: DynamoDBで有効な型シグネチャ
    """

    tp = type(val)
    if tp is int:
        return 'N'
    if tp is set:
        t = list(val)[0]
        return _to_type_string(t) + 'S'
    if tp is str:
        return 'S'
    if tp is bytes:
        return 'B'
    if tp is dict:
        return 'M'
    if tp is list:
        return 'L'
    if tp is bool:
        return 'BOOL'
    else:
        return '__extra__'


def primal_serializer(val):
    """
    基本型のシリアライザ。DynamoDBからPythonに変換する

    :param val: DynamoDBドキュメント
    :return: pythonでの値
    """
    s = _to_type_string(val)

    if s in 'L':
        return {s: [primal_serializer(ss) for ss in val]}
    if s in {'SS', 'NS', 'BS'}:
        return {s: [ss for ss in val]}
    if s == 'M':
        return {s: {k: primal_serializer(v) for k, v in val.items()}}
    if s in {'B', 'BOOL'}:
        return {s: val}
    if s == '__extra__':
        if type(val) is UUID:
            return primal_serializer(val.hex)
        if type(val) is datetime:
            return primal_serializer(val.timestamp())
        else:
            raise Exception(f'invalid type {val} {type(val)}')

    return {s: str(val)}



from typing import Generic, TypeVar, Type, Optional

from hatsudenki.packages.expression.update import UpdateExpression
from hatsudenki.packages.field import BaseHatsudenkiField
from hatsudenki.packages.marked import MarkedObjectDict

T = TypeVar('T')


class DictMapField(BaseHatsudenkiField[MarkedObjectDict[T]], Generic[T]):
    PythonType = dict
    TypeStr = 'M'
    TypeName = 'DictMap'

    IsScalar = False

    def __init__(self, value_type: Type[T], *, name=None, default=None, ttl=False):
        super().__init__(name=name, default=default, ttl=ttl)
        self.value_type = value_type

    def get_data_from_dict(self, v: dict, table=None) -> MarkedObjectDict[T]:
        return super().get_data_from_dict(v, table)

    def get_data(self, val, table=None) -> Optional[T]:
        if self.is_empty(val):
            return MarkedObjectDict(self.name, self.value_type, table)

        if isinstance(val, self.value_type):
            return MarkedObjectDict(self.name, self.value_type, table, val)
        if isinstance(val, MarkedObjectDict):
            return val

        if type(val) is dict:
            return MarkedObjectDict(self.name, self.value_type, table, val)

        raise Exception(f'invalid value given. {val}')

    @classmethod
    def is_empty(cls, value: MarkedObjectDict[T]):
        if value is None:
            return True
        return value.is_empty()

    def serialize(self, value: MarkedObjectDict[T], table=None):
        if self.is_empty(value):
            return None

        return value.serialized_value

    def deserialize(self, value, table=None) -> MarkedObjectDict[T]:
        if value is None:
            return MarkedObjectDict(self.name, self.value_type, table, None)

        return MarkedObjectDict.deserialize(self.name, self.value_type, value, table)

    def build_update_expression(self, upd: UpdateExpression, now_value):
        now_value.build_update_expression(upd)



from typing import Type, TypeVar, Iterator

from hatsudenki.packages.cache.base.memory.solo import CacheBaseTableSolo

T = TypeVar('T')


class StaticCacheBaseTableSolo(CacheBaseTableSolo):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        pass

    @classmethod
    def get(cls: Type[T], hash_key, range_key=None) -> T:
        """
        HASHキーを指定して一件取得

        :param hash_key: ハッシュキー値
        :param range_key: レンジキー値（Soloテーブルはこの値は必ずNoneになる）
        :return: 条件にマッチしたレコード
        """

        return cls._get(hash_key, range_key)

    @classmethod
    def get_by_cursor(cls: Type[T], cursor: str) -> T:
        """
        カーソル文字列から一件取得

        :param cursor: カーソル文字列
        :return: 条件にマッチしたレコード
        """
        return cls._get_by_cursor(cursor)

    @classmethod
    def check_exist_hash(cls, hash_key):
        """
        指定されたハッシュキーのレコードがあるか判定する

        :param hash_key: 検索対象値
        :return: bool
        """
        return cls._check_exist_hash(hash_key)

    @classmethod
    def iter(cls: Type[T]) -> Iterator[T]:
        """
        全件列挙イテレータ

        :return: iter
        """

        return cls._iter()



from datetime import datetime
from logging import getLogger
from typing import Optional, Type, TypeVar
from uuid import UUID

from hatsudenki.packages.field.base import BaseHatsudenkiField

_logger = getLogger(__name__)

T = TypeVar('T')


class UUIDField(BaseHatsudenkiField[UUID]):
    PythonType = UUID

    TypeStr = 'S'
    TypeName = 'UUID'

    def serialize(self, value, table=None):
        if type(value) is str:
            u = UUID(hex=value)
            return {'S': u.hex}

        if self.__class__.is_empty(value):
            return None
        return {'S': value.hex}

    def deserialize(self, value, table=None):
        if value is None:
            return self.default_value
        return UUID(hex=value['S'])

    def get_data(self, val, parent=None) -> Optional[UUID]:
        if val is None:
            return None

        if type(val) is UUID:
            return val
        if type(val) is str:
            return UUID(hex=val)

    def to_string(self, val: UUID):
        return str(val.hex)


class DateField(BaseHatsudenkiField[datetime]):
    TypeStr = 'N'
    PythonType = datetime
    TypeName = 'Date'

    def serialize(self, value, table=None):
        if type(value) is str:
            d = datetime.fromtimestamp(int(value))
            return {'N': str(int(d.timestamp()))}

        if self.is_empty(value):
            return None
        return {'N': str(int(value.timestamp()))}

    def deserialize(self, value, table=None):
        if value is None:
            return self.default_value
        return datetime.fromtimestamp(int(value['N']))

    def to_string(self, val: datetime):
        return str(int(val.timestamp()))

    def get_data(self, val, table=None):
        if type(val) is datetime:
            return val
        if type(val) is int:
            return datetime.fromtimestamp(val)
        if type(val) is str:
            return datetime.fromisoformat(val)


class CreateDateField(DateField):
    pass


class UpdateDateField(DateField):
    pass


class EnumField(BaseHatsudenkiField[int]):
    TypeStr = 'N'
    PythonType = int
    TypeName = 'Enum'

    def __init__(self, to: Type[T], **kwargs):
        super().__init__(**kwargs)
        self.to = to

    def serialize(self, value: T, table=None):
        if type(value) == self.to:
            return {'N': str(value.value)}
        elif type(value) == int:
            return {'N': str(value)}
        elif self.is_empty(value):
            return None

    def deserialize(self, value, table=None):
        try:
            return self.to(int(value['N']))
        except ValueError as e:
            _logger.warning(f'Enum value is not exist. table: {table}, value: {int(value["N"])}')
            return None

    def to_string(self, val):
        return str(int(val.value))

    def get_data(self, val, table=None):
        if type(val) == self.to:
            return val
        elif type(val) == int:
            return self.to(val)
        elif self.is_empty(val):
            return None



from datetime import datetime


class DateResolver(object):
    def get_now(self):
        raise NotImplementedError()


class DefaultDateResolver(DateResolver):
    def get_now(self):
        return datetime.now()


class DateManager:
    _resolver = DateResolver()

    @classmethod
    def set_resolver(cls, resolver: DateResolver):
        cls._resolver = resolver

    @classmethod
    def get_now(cls):
        return cls._resolver.get_now()



from pathlib import Path
from typing import List, Dict

import openpyxl
from openpyxl.worksheet.worksheet import Worksheet

from hatsudenki.packages.command.stdout.output import ToolOutput


class ImportMasterExcelSheet(object):
    def __init__(self, book: 'ImportMasterExcel', sheet):
        self.book = book
        self.sheet: Worksheet = sheet

    def add_table(self, table_data: List[dict]):
        ToolOutput.anchor(f'{self.book.path.name} {self.sheet.title}にレストア')
        r = self.sheet.rows
        headers = next(r)
        idxes = [h.value for h in headers]

        self.sheet.delete_rows(2, self.sheet.max_row - 1)

        for d in table_data:
            dd = []
            for idx in idxes:
                if idx == 'TAG':
                    # TAGは出力されないので強制的に0入れる
                    dd.append(0)
                    continue
                dd.append(d.get(idx, None))
            self.sheet.append(dd)
        ToolOutput.pop('OK')


class ImportMasterExcel(object):
    def __init__(self, path: Path):
        self.path = path
        self.book: openpyxl.Workbook = None
        self.sheets: Dict[str, ImportMasterExcelSheet] = {}
        self.load()

    def load(self):
        if self.book is None:
            self.book = openpyxl.load_workbook(str(self.path), data_only=False)
            self.sheets = {sn: ImportMasterExcelSheet(self, self.book[sn]) for sn in self.book.sheetnames}

    def save(self):
        self.book.save(self.path)



from datetime import datetime
from pathlib import Path
from typing import Dict, Tuple

from openpyxl.worksheet.worksheet import Worksheet

from hatsudenki.packages.command.excel.excel_data import ExcelData
from hatsudenki.packages.command.master.column import MasterColumn, ColumnRelation, ColumnChose
from hatsudenki.packages.command.master.table import MasterTable
from hatsudenki.packages.command.master.tag.loader import MasterTag
from hatsudenki.packages.command.stdout.output import ToolOutput
from hatsudenki.packages.command.util.base_info import BaseInfo


class MasterExcelData(ExcelData):
    """
    企画入力マスターデータExcel
    """

    def __init__(self, file_path: Path):
        super().__init__(file_path)
        self.load()

    def iter_table_sheet(self):
        return (sheet for sheet in self.book.worksheets)

    def iter_formula_table_sheet(self):
        return (sheet for sheet in self.formula_book.worksheets)


class MasterExcelSheet(object):
    """
    企画入力マスターシート
    """

    def __init__(self, book: 'MasterExcel', sheet: Worksheet, formula_sheet: Worksheet, table: MasterTable, loader):
        from hatsudenki.packages.command.master.exporter.loader import MasterExcelLoader
        self.sheet = sheet
        self.formula_sheet = formula_sheet
        self.table = table
        self.loader: MasterExcelLoader = loader
        ToolOutput.out(f'[MasterExcel]Sheet準備 {self.sheet.title} = {table.table_name}')
        self.book = book

        self.resolved_headers = self._resolve_headers()
        self.datas: Dict[str, Dict[str, Tuple[MasterColumn, any]]] = self._resolve()

    @property
    def reference_label(self):
        return self.table.reference_column

    @property
    def cursor_labels(self):
        return self.table.cursor_labels

    @property
    def has_range(self):
        return self.table.range_key is not None

    @property
    def except_label(self):
        return f'{self.book.data.name} {self.sheet.title} "{self.table.full_path}:0"'

    def _resolve_headers(self):
        h = next(self.formula_sheet.rows)
        ret = {}
        for cell in h:
            c = next((c for c in self.table.columns.values() if cell.value == c.excel_header_name), None)
            ret[cell.value] = c

        return ret

    def _check_relation(self, table: MasterTable, val, header: str = None):
        excel = self.loader.get_by_excel_name(table.excel_name + '.xlsx')
        s = excel.get_sheet_by_name(table.excel_sheet_name)
        d = s.find_row(val)
        if d is None:
            target_info = f'{table.label} "{table.rel_path}:0"'
            raise Exception(f'参照解決に失敗 {self.except_label} {header} {val}\n{target_info}')

        return val

    def _check_relation_by_column(self, column: ColumnRelation, val):
        t = column.get_target_table()
        return self._check_relation(t, val, column.excel_raw_header_name)

    def _resolve_chose(self, column: ColumnChose, sel, val):
        """
        choseカラムの解決

        :param column: カラムインスタンス
        :param sel: セレクタの値
        :param val: ID
        :return:
        """
        v = column.resolve_value(sel, val, is_raw=False)
        if isinstance(v, MasterTable):
            # masterの場合はリレーション先が存在するかチェックする
            # enumは解決時にチェックされているので不要
            self._check_relation(v, val, column.excel_raw_header_name)
        # strで固定しないとMessagePackに変換できない
        return str(v)

    def _find_header_index(self, key_str: str):
        return next((idx for idx, key in enumerate(self.resolved_headers.keys()) if key == key_str), None)

    def _check_tag_level(self, tag_level: MasterTag):

        if tag_level.is_debug and not self.is_enable_debug:
            # デバッグタグはデバッグモードが有効でないときは処理しない
            return False

        if self.out_tag_level.is_only:
            # Onlyが設定されている場合は完全一致のみ
            if tag_level.name == self.out_tag_level.name:
                return True
            # と見せかけてincludeに入ってるやつもOK
            if tag_level.name in self.out_tag_level.include:
                return True
            return False

        if tag_level.is_only:
            # Onlyは指定タグに設定されたときしか出力されない
            return False

        # TAGレベル判定
        if tag_level.level > self.out_tag_level.level:
            return False
        return True

    @property
    def out_tag_level(self):
        return self.loader.out_tag

    @property
    def is_enable_debug(self):
        return self.loader.is_debug

    @staticmethod
    def _conv_datetime(value: datetime):
        # msが存在しない場合はそのまま返して良い
        if value.microsecond == 0:
            return value
        return datetime.fromtimestamp(round(value.timestamp(), -1))

    def _resolve(self):
        rows = self.sheet.rows
        next(rows)
        tag_idx = self._find_header_index('TAG')

        ref_idxes = [self._find_header_index(cursor_column.excel_header_name) for cursor_column in self.cursor_labels]

        datas = {}
        for row in rows:
            ret = {}

            # TAGレベル判定
            tag_level = str(row[tag_idx].value)
            tag_info = self.loader.resolve_tag(tag_level)

            if row[ref_idxes[0]].value is None or not self._check_tag_level(tag_info):
                continue

            if tag_info is None:
                raise Exception(f'無効なタグ指定 {tag_level}')

            # 既存データの検索
            hash = row[ref_idxes[0]].value
            already = datas.get(hash)
            if already and self.has_range:
                range = row[ref_idxes[1]].value
                already = already.get(range)

            # TAGレベルが現在データよりも低い場合は上書きしない
            if already:
                at = self.loader.resolve_tag(str(already['__TAG__'][1]))
                if at.level > tag_info.level:
                    # 前データのほうが強いのでスキップ
                    continue
                elif at.level == tag_info.level:
                    if at.is_debug == tag_info.is_debug:
                        # デバッグ属性が同一かつレベルが同じ＝Excel内でキーが重複している
                        if self.has_range:
                            raise Exception(
                                f'キーが重複しています {self.book.data.name} {self.sheet.title} {self.table.hash_key.excel_raw_header_name} {hash} - {self.table.range_key.excel_raw_header_name} {range} "{self.table.full_path}:0"')
                        else:
                            raise Exception(
                                f'キーが重複しています {self.book.data.name} {self.sheet.title} {self.table.hash_key.excel_raw_header_name} {hash}"{self.table.full_path}:0"')

                    if not tag_info.is_debug:
                        # すでにデバッグデータで上書きされているので処理しない
                        continue

            for header_key, col in zip(self.resolved_headers.keys(), row):
                if header_key == 'TAG':
                    ret['__TAG__'] = None, col.value
                    continue

                header = self.resolved_headers[header_key]
                if header is None:
                    continue

                # エクセルからロードした値のチェック
                value = col.value
                # 値がdatetime型の時
                if type(value) is datetime:
                    # 値のコンバートが必要な場合がある
                    value = self._conv_datetime(value)
                ret[header.python_name] = header, value

            if self.has_range:
                if hash not in datas:
                    datas[hash] = {ret[self.table.range_key.python_name][1]: ret}
                else:
                    datas[hash][ret[self.table.range_key.python_name][1]] = ret
            else:
                datas[hash] = ret

        return datas

    def iter(self):
        """
        1行のデータのイテレータ

        :return: dict
        """

        def _(k, c, v, ret: dict, hm: dict):
            if c is None:
                return
            if isinstance(c, ColumnRelation):
                ret[k] = self._check_relation_by_column(c, v)
            elif isinstance(c, ColumnChose):
                ret[k] = self._resolve_chose(c, hm[c.selector][1], v)
            else:
                ret[k] = c.generate_value(v)

        for hash_map in self.datas.values():
            if self.has_range:
                for mapped_data in hash_map.values():
                    ret = {}
                    for k, (c, v) in mapped_data.items():
                        _(k, c, v, ret, mapped_data)
                    yield ret
            else:
                ret = {}
                for k, (c, v) in hash_map.items():
                    _(k, c, v, ret, hash_map)
                yield ret

    def find_row(self, value):
        return self.datas.get(value)


class MasterExcel(BaseInfo):
    def __init__(self, base_path: Path, full_path: Path, table_data: Dict[str, MasterTable], loader, *args,
                 **kwargs):
        from hatsudenki.packages.command.master.exporter.loader import MasterExcelLoader
        super().__init__(base_path, full_path, *args, **kwargs)
        self.data = MasterExcelData(Path(full_path))
        self.table = table_data
        self.loader: MasterExcelLoader = loader
        self.table_sheets: Dict[str, MasterExcelSheet] = {}
        ToolOutput.anchor(f'{self.data.name} をよみこみます')
        for sheet in self.data.iter_table_sheet():
            if sheet.title not in self.table:
                continue
            self.table_sheets[sheet.title] = self._create_sheet(sheet)
        ToolOutput.pop('OK')

    def iter_table_sheet(self):
        return (sheet for sheet in self.table_sheets.values())

    def get_sheet_by_name(self, sheet_name: str):
        return self.table_sheets[sheet_name]

    def _create_sheet(self, sheet):
        return MasterExcelSheet(self, sheet, self.data.formula_book[sheet.title],
                                self.table[sheet.title], self.loader)



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Tuple, Iterable

from hatsudenki.packages.command.hatsudenki.data import HatsudenkiData
from hatsudenki.packages.command.renderer.base import RenderUnit, FileRenderer
from hatsudenki.packages.command.stdout.output import IndentString


class TableRenderUnit(RenderUnit[HatsudenkiData]):
    """
    DynamoDBTableレンダラ
    """

    def __init__(self, data: HatsudenkiData):
        """
        イニシャライザ

        :param data: Dataインスタンス
        """
        super().__init__(data)
        # indexをパースしておく
        self.index_info, self.gsi_units = self.data.parse_index()

    @staticmethod
    def _render_opt_str(opt: dict):
        """
        dictをpythonのnamedparam形式文字列に変換する
        {xxx:yyy} -> xxx=yyy

        :param opt: 対象dict
        :return: str
        """
        return ', '.join(f'{k}={json.dumps(v) if json.dumps(v) != "null" else None}' for k, v in opt.items())

    def render_meta(self):
        """
        Metaクラスの出力

        :return: IndexStringインスタンス
        """
        table = self.data
        # hashキー及びrangeキーを取得
        hk = table.hash_key
        rk = table.range_key

        # 単一か子かによって継承元が変わる
        is_alone = table.is_alone
        if is_alone:
            # 単一テーブルの場合は普通に親から継承
            meta = IndentString('class Meta(SoloHatsudenkiTable.Meta):')
        else:
            # 子テーブルの場合はベースに加え、親テーブルのMetaも継承
            meta = IndentString(f'class Meta(ChildMultiHatsudenkiTable.Meta, {table.parent_table.class_name}.Meta):')

        # 共通属性
        meta.add(f'label = "{table.label}"')
        meta.add(f'is_root = {is_alone}')
        meta.add(f'table_name = "{table.table_name}"')

        # 単一テーブルか否かで処理を分岐
        if is_alone:
            # コレクション名
            meta.add(f'collection_name = "{table.collection_name}"')
            # キー情報
            opt = {
                'hash_key': hk.name
            }
            if rk is not None:
                # rangeキーが存在する場合は追加
                opt['range_key'] = rk.name

            # キャパシティユニット
            cap_unit = table.capacity_units
            opt['read_cap'] = cap_unit['read']
            opt['write_cap'] = cap_unit['write']
            # インデックス
            meta.add(f'primary_index = PrimaryIndex(name=None, {self._render_opt_str(opt)})')
        else:
            # 個別にクラスを出力する必要がある場合
            if rk:
                meta.add(rk.class_string)

            # TAGネーム
            meta.add(f'tag_name = "{table.tag_name}"')
            if rk is not None:
                # aliasキー
                meta.add(f'key_alias_name = "{rk.name}"')
                # aliasキータイプ
                meta.add(f'key_alias_type = {rk.def_python_field_class({})}')
        # 空行
        meta.blank_line()
        return meta

    def render_index(self):
        """
        インデックス情報の出力

        :return: IndexString
        """
        table = self.data
        # 単一か子かによって継承元が変わる
        is_alone = table.is_alone
        if is_alone:
            # 個別
            lsi_class_str = IndentString('class LSIndex:')
            gsi_class_str = IndentString('class GSIndex:')
        else:
            # 子
            pn = table.parent_table.class_name
            lsi_class_str = IndentString(f'class LSIndex({pn}.LSIndex):')
            gsi_class_str = IndentString(f'class GSIndex({pn}.LSIndex):')

        # インデックスは複数件あるのでループ
        for idx in table.indexes:
            opt = {}
            opt['projection_keys'] = idx.get('projection', [])
            # LSI or GSI
            if idx['type'] == 'global':
                # GSIはhashキーとrangeキーとキャパシティユニット
                opt['hash_key'] = idx['hash']
                # インデックス名はgsi__{hash}__{range}
                opt['name'] = f'gsi_{idx["hash"]}'
                if 'range' in idx:
                    opt['range_key'] = idx['range']
                    opt['name'] += '__' + idx['range']
                # キャパシティユニット
                caps = idx.get('capacity_units', {})
                if 'capacity_units' in caps:
                    c = caps['capacity_units']
                    opt['read_cap'] = c['read']
                    opt['write_cap'] = c['write']
                else:
                    opt['read_cap'] = 1
                    opt['write_cap'] = 1
                s = ', '.join((f'{k}={json.dumps(v)}' for k, v in opt.items()))
                gsi_class_str.add(f'{opt["name"]} = GSI({s})')
            else:
                # LSIはrangeキーのみ
                opt['name'] = f'lsi_{idx["range"]}'
                opt['range_key'] = idx['range']
                # 処理の都合上PrimaryのHashキーを便宜上のhashキーとして登録する
                if is_alone:
                    opt['hash_key'] = table.hash_key.name
                else:
                    # 子テーブルの場合はPrimaryキーを親テーブルから引く
                    opt['hash_key'] = table.parent_table.hash_key.name
                s = ', '.join((f'{k}={json.dumps(v)}' for k, v in opt.items()))
                lsi_class_str.add(f'{opt["name"]} = LSI({s})')

        # LSIが一つもない場合はpass
        if lsi_class_str.line_num is 1:
            lsi_class_str.add('pass')

        # GSIが一つもない場合はpass
        if gsi_class_str.line_num is 1:
            gsi_class_str.add('pass')
        body = IndentString()

        body.add(lsi_class_str, gsi_class_str)

        return body

    def render_fields(self):
        """
        Fieldクラスの出力

        :return: IndexString
        """
        table = self.data
        # 継承元の決定
        is_alone = table.is_alone
        if table.is_alone:
            body = IndentString('class Field(SoloHatsudenkiTable.Field):')
        else:
            body = IndentString(f'class Field({table.parent_table.class_name}.Field):')

        attr_num = 0
        for key, attr in table.attributes.items():
            if attr.is_alias_key:
                # aliasキーなので実際のフィールドとしては存在しない
                continue
            body.add(attr.class_string)
            body.add(attr.gen_def_str())

            attr_num += 1

        # 属性が一つもない場合はpass
        if attr_num is 0:
            body.add('pass')
        return body

    def render_init(self):
        """
        イニシャライザの出力

        :return: IndexString
        """
        table = self.data
        # キー情報の取得
        hk = table.hash_key
        rk = table.range_key
        keys = [k.name for k in [hk, rk] if k is not None]
        # hashキーとrangeキーだけ明示的にパラメータとして定義する
        o = ', '.join(
            [f'{k}: {table.attributes[k].python_init_type_str} = None' for k in keys])
        if len(o) > 0:
            o = f'{o}, '

        body = IndentString(f'def __init__(self, {o}**kwargs):')
        body.add('super().__init__(**kwargs)')
        # 記述を短くするためにFieldクラスの参照を保持する
        body.add('ft = self.__class__.Field')

        is_alone = table.is_alone
        if not is_alone and rk is None:
            # 子テーブル且つレンジキーがない場合は単一子テーブルとなる
            parent = table.parent_table
            # 単一子テーブルのrangeキーの値は常にtagで固定
            body.add(f'self.{parent.range_key.name} = "{table.tag_name}"')

        # 各フィールドの初期化処理
        for key, attr in table.attributes.items():
            body.add(attr.gen_init_str())

        return body

    def render_resolver(self):
        """
        リゾルバの出力

        :return: IndexString
        """
        # 今の所aliasキーのリゾルバしか無い
        body = IndentString()
        for k, attr in self.data.attributes.items():
            b = attr.resolver_string
            if b:
                body.add(b)

        return body

    def render(self):
        """
        出力

        :return: IndentString
        """
        table = self.data

        # 継承元の決定
        if table.is_alone:
            # solo or multi
            if table.range_key is not None:
                parent_cls_name = 'MultiHatsudenkiTable'
            else:
                parent_cls_name = 'SoloHatsudenkiTable'
        else:
            # solo or multi
            if table.range_key:
                parent_cls_name = f'{table.parent_table.class_name}, ChildMultiHatsudenkiTable'
            else:
                parent_cls_name = f'{table.parent_table.class_name}, ChildSoloHatsudenkiTable'

        # クラス定義
        cls = IndentString(f'class {table.class_name}({parent_cls_name}):')

        # Metaクラス
        cls.add(self.render_meta())

        cls.blank_line()

        cls.add(self.render_fields())
        cls.blank_line()

        # Index
        cls.add(self.render_index())
        cls.blank_line()
        # init
        cls.add(self.render_init())
        cls.blank_line()
        # resolver
        cls.add(self.render_resolver())

        cls.blank_line()

        return cls


class TableFileRenderer(FileRenderer):
    """
    ファイルレンダラ
    """

    def __init__(self, table_iterator: Iterable[Tuple[Path, HatsudenkiData]], out_module_name: str):
        """
        イニシャライザ

        :param table_iterator: Dataのイテレータ
        """
        super().__init__()

        # rootとchildに切り分ける
        d = {'root': [], 'child': []}
        [d['root' if v.is_alone else 'child'].append(v) for k, v in table_iterator]

        # rootから先に書く
        [self.add_unit(TableRenderUnit(t)) for t in d['root']]
        [self.add_unit(TableRenderUnit(t)) for t in d['child']]

        self.out_module_name = out_module_name.replace(os.sep, '.')

    def render_header(self) -> IndentString:
        """
        ヘッダ出力

        :return: IndentString
        """
        # インポートするモジュール群
        d = IndentString()
        l = [
            'from __future__ import annotations',
            'import uuid',
            f'from {self.out_module_name} import masters',
            'from datetime import datetime',
            f'from {self.out_module_name} import def_enum',
            'from hatsudenki.packages import field',
            'from hatsudenki.packages.marked import MarkedObject, Markable, MarkedObjectWithIndex',
            'from hatsudenki.packages.table.index import PrimaryIndex, LSI, GSI',
            'from hatsudenki.packages.table.multi import MultiHatsudenkiTable',
            'from hatsudenki.packages.table.solo import SoloHatsudenkiTable',
            'from hatsudenki.packages.table.child import ChildMultiHatsudenkiTable',
            'from hatsudenki.packages.table.child_solo import ChildSoloHatsudenkiTable',
        ]
        d.add(*l)
        d.blank_line(2)

        d.indent('def table_setup():')
        d.add('from hatsudenki.packages.manager.table import TableManager')
        d.add(
            'print(f"load all tables OK. collections={TableManager.get_collection_num()}  tables={TableManager.get_table_num()}")')
        d.blank_line(2)

        return d



from logging import getLogger
from typing import TypeVar, Type

import dill

T = TypeVar('T')
_logger = getLogger(__name__)


class StoreCacheBaseTable(object):
    class Meta:
        table_name: str = ''

    def serialize(self):
        return dill.dumps(self)

    @classmethod
    def deserialize(cls: Type[T], b: any) -> T:
        return dill.loads(b)

    @classmethod
    def _gen_key(cls, hash_value, range_value):
        k = f'{cls.get_table_name()}:{str(hash_value)}'

        if range_value:
            k += '--' + str(range_value)
        return k

    @classmethod
    def get_table_name(cls):
        return cls.Meta.table_name

    @classmethod
    async def build(cls, hash_value, range_value=None):
        pass

    @classmethod
    async def _store(cls, hash_value, range_value, data: bytes):
        pass

    @classmethod
    async def get(cls: Type[T], hash_value, range_value) -> T:
        v = await cls._get(hash_value, range_value)
        if v is None:
            _logger.info(f'setup {cls.Meta.table_name}')
            r = await cls.build(hash_value, range_value)
            await cls._store(hash_value, range_value, r.serialize())
            return r
        return cls.deserialize(v)

    @classmethod
    async def _get(cls, hash_key, range_key) -> bytes:
        pass



from hatsudenki.packages.field import T
from hatsudenki.packages.field.base import BaseHatsudenkiField


class StringField(BaseHatsudenkiField[str]):
    """
    文字列フィールドクラス
    """
    PythonType = str
    TypeStr = 'S'
    TypeName = 'String'

    @classmethod
    def is_empty(cls, value):
        return value is None or len(value) is 0

    def deserialize(self, value, table=None):
        if value is None:
            return self._default_value
        return value['S']

    def serialize(self, value: T, table=None):
        if value is None or len(value) is 0:
            return None

        return {'S': value}


class NumberField(BaseHatsudenkiField[int]):
    """
    数値フィールドクラス。符号はどっちも入る。少数は駄目
    """
    PythonType = int
    TypeStr = 'N'
    TypeName = 'Integer'

    # @classmethod
    # def add_expression(cls, name: str, new_value, old_value, update: UpdateExpression):
    #     if old_value is None:
    #         update.add()
    #     update.increment(name, cls.serialize(new_value - old_value, None), raw=True)

    def deserialize(self, value, table=None):
        if value is None:
            return self._default_value

        return int(value['N'])

    def serialize(self, value: T, table=None):
        if value is None:
            return None

        return {'N': f'{value}'}


class BinaryField(BaseHatsudenkiField[bytes]):
    """
    バイナリフィールドクラス
    """
    PythonType = bytes
    TypeStr = 'B'
    TypeName = 'Binary'

    @classmethod
    def is_empty(cls, value):
        return value is None or len(value) is 0

    def deserialize(self, value, table=None):
        if value is None:
            return self._default_value

        return value['B']

    def serialize(self, value: T, table=None):
        if value is None or len(value) is 0:
            return None

        return {'B': value}


class BoolField(BaseHatsudenkiField[bool]):
    """
    ブーリアンフィールドクラス
    """
    PythonType = bool
    TypeStr = 'BOOL'
    TypeName = 'Boolean'

    def deserialize(self, value, table=None):
        if value is None:
            return self._default_value

        return value['BOOL']

    def serialize(self, value: T, table=None):
        if value is None:
            return None

        return {'BOOL': value}



from typing import Generic, Type, Union, List, Iterator, TypeVar

from hatsudenki.packages.expression.update import UpdateExpression
from hatsudenki.packages.marked import Markable

T = TypeVar('T')


class MarkedObjectList(Generic[T]):
    """
    MarkedObjectのリスト
    """

    def __init__(self, name: str, target_class: Type[T], parent: Markable,
                 items: Union[List[T], 'MarkedObjectList[T]', None] = None):
        """
        イニシャライザ

        :param name: 名前
        :param target_class: 格納可能なインスタンスタイプ
        :param parent: 親オブジェクト
        :param items: 初期値
        """
        if isinstance(items, MarkedObjectList):
            self._list: List[T] = items._list
        elif type(items) is list:
            self._list: List[T] = items
        else:
            self._list: List[T] = []
        self.name = name
        self._target_class: Type[T] = target_class
        self._mark = set()
        self._appended = []
        self._parent = parent
        self._mark_replace = False

    def __len__(self):
        return len(self._list)

    def append(self, data: T):
        """
        末尾に追加

        :param data: 追加するデータ
        :return: None
        """
        data._parent = self
        i = len(self._list)
        data.name = i
        # data.set_appended()
        self._list.append(data)
        self.append_mark(data)

    def replace(self, data: List[T]):
        """
        置き換え

        :param data:
        :return:
        """
        self.flush()
        self.set_list(data)
        self.replace_mark()

    def __getitem__(self, item: int) -> T:
        """
        [x]形式でアクセス

        :param item: 添字
        :return: T
        """
        return self._list[item]

    def __setitem__(self, key: int, value: T):
        """
        [x]形式でセット。変更もマークする

        :param key: 添字
        :param value: セットする値
        :return: None
        """
        if value is not None:
            # 親情報をセット
            value._parent = self._parent
            value.name = key

        self._list[key] = value
        # 更新をマーク
        self.modify_mark(key)

    def __iter__(self) -> Iterator[T]:
        """
        イテレータ

        :return: iterator
        """
        return self._list.__iter__()

    def create(self, **kwargs) -> T:
        """
        格納可能なインスタンスを生成して返す（追加はしない）

        :param kwargs: 初期値
        :return: 生成されたインスタンス
        """
        return self._target_class(name=None, parent=None, **kwargs)

    def modify_mark(self, key):
        """
        更新をマーク
        :param key: 更新されたキー
        :return: None
        """

        if self._parent:
            # 親が設定されていた場合は親にも伝搬する
            self._parent.modify_mark(self.name)
        self._mark.add(key)

    def append_mark(self, data: T):
        """
        要素追加をマーク

        :param data: 追加された要素
        :return:
        """
        if self._parent:
            # 親が設定されていた場合は親に更新を伝搬する
            self._parent.modify_mark(self.name)
        # 追加された要素リストに追加
        self._appended.append(data)

    def replace_mark(self):
        """
        置き換えをマーク
        ※重いよ

        :return: None
        """
        if self._parent:
            # 親が設定されていた場合は親に更新を伝搬する
            self._parent.modify_mark(self.name)
        self._mark_replace = True

    def flush(self):
        """
        変更情報をクリア

        :return: None
        """
        self._mark.clear()
        self._appended.clear()
        self._mark_replace = False

    def set_list(self, d: List[T]):
        """
        リストを直接セットする。この処理は更新をマークしない。

        :param d: セットするリスト
        :return:
        """
        self._list = d

    @classmethod
    def serialize(cls, data: 'MarkedObjectList[MarkedObject]'):
        """
        DynamoDBへシリアライズ

        :param data: 対象データ
        :return: DynamoDB dict
        """
        ret = [d.serialized_value for d in data]
        return {'L': ret}

    @property
    def serialized_value(self):
        """
        DynamoDB dictにシリアライズされた値

        :return: DynamoDB dict
        """
        return self.__class__.serialize(self)

    @classmethod
    def deserialize(cls, name, target_class: Type[T], data: dict, parent: Markable):
        """
        pythonにデシリアライズ
        :param name: 名前
        :param target_class: 格納対象クラスタイプ
        :param data: dynamoDB dict
        :param parent: 親オブジェクト
        :return: 自分自身をインスタンス化したものy
        """
        der = target_class.deserialize
        ret = cls(name, target_class, parent)
        ret.set_list([der(idx, i, ret) for idx, i in enumerate(data['L'])])
        return ret

    def build_update_expression(self, upd: UpdateExpression):
        """
        更新箇所を考慮したクエリをビルド。
        追加と更新が同時に行われた場合は例外が発生する

        :param upd: UpdateExpressionインスタンス
        :return: None
        """

        if self._mark_replace:
            # replaceがマークされているときはまるっと上書き。Replaceマークはすべてのマークより優先される
            upd.set(
                self.name,
                {'L': [u.serialized_value for u in self._list]},
                raw=True
            )
            return

        # 一つの要素に対し二度オペレーションを行うことはできない(DynamoDBの制限)
        # if len(self._appended) > 0 and len(self._mark) > 0:
        #     raise Exception('追加と更新を同時に行うことはできません')

        # 追加された要素がある場合はlist_appendを発行する
        if len(self._appended) > 0:
            upd.list_append(self.name, {'L': [u.serialized_value for u in self._appended]}, raw=True)
            # return

        # 更新キーを考慮して小要素を直接セットするクエリをビルド
        for update_key in self._mark:
            now_value = self[update_key]
            fc = self._target_class

            n = f'{self.name}[{update_key}]'
            if fc.is_empty(now_value):
                # 空になった場合はremoveを発行する
                upd.remove_raw_key(n)
                continue

            now_value.build_update_expression(n, upd)



from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, List

from hatsudenki.packages.command.files import yaml_load


@dataclass
class MasterTag:
    name: str
    data: dict
    level: int
    include: List[str]
    is_debug: bool

    @property
    def is_only(self):
        return self.data.get('only')


class MasterTagLoader:
    def __init__(self, yaml_path: Path):
        self.yaml_path = yaml_path
        self._data: Optional[Dict[str, MasterTag]] = None

    def load(self):
        if self._data is None:
            self._data = {}
            d = yaml_load(self.yaml_path)

            for idx, t in enumerate(d):
                if 'debug_only' in t and t['debug_only'] is True:
                    self._data[t['name']] = MasterTag(name=t['name'], data=t, level=idx,
                                                      include=t.get('include', []), is_debug=True)
                else:
                    self._data[t['name']] = MasterTag(name=t['name'], data=t, level=idx, include=t.get('include', []),
                                                      is_debug=False)
                    self._data[t['name'] + '_debug'] = MasterTag(name=t['name'] + '_DEBUG', data=t, level=idx,
                                                                 include=t.get('include', []), is_debug=True)

    def get_tag(self, tag_name: str):
        return self._data.get(tag_name, None)



from typing import Generic, TypeVar, Type, Optional, Dict, Iterator

from hatsudenki.packages.expression.update import UpdateExpression
from hatsudenki.packages.marked import MarkedObject, Markable

T = TypeVar('T')


class MarkedObjectWithIndex(MarkedObject):
    """
    Hashキーを指定できるMarkedObject
    """

    class Meta:
        # ハッシュキーの名前
        hash_key = ''

    @classmethod
    def get_hash_key_name(cls):
        """
        ハッシュキーの名前を取得

        :return: str
        """
        return cls.Meta.hash_key

    def get_hash_value(self):
        """
        ハッシュキーの値を取得

        :return: str
        """
        return str(getattr(self, self.__class__.get_hash_key_name()))

    def __init__(self, name, parent: Markable, **kwargs):
        super().__init__(name, parent, **kwargs)
        self._is_appended = True

    def flush(self):
        super().flush()
        # 値が空のときにappendフラグを追ってしまうと新規追加を検知できない
        # そのままupdateされてもmarkが空なのでどのみちクエリは発行されないのでOK
        if not self.__class__.is_empty(self):
            self._is_appended = False

    def build_update_expression(self, parent_name, upd: UpdateExpression):
        # 新規追加の際は無条件に全部
        if self._is_appended:
            upd.set(parent_name, self.serialized_value, raw=True)
        else:
            super().build_update_expression(parent_name, upd)


class MarkedObjectDict(Generic[T]):
    def __init__(self, name: str, target_class: Type[T], parent: Optional[Markable], item: Optional[dict] = None):
        """
        イニシャライザ

        :param name: フィールド名前
        :param target_class: 子として抱えるクラスタイプ
        :param parent: 親Markable
        :param item: 要素
        """
        self.name = name
        self._target_class = target_class
        self._mark = set()
        self._item: Dict[any, T] = {}
        if item:
            for k, v in item.items():
                self._item[k] = target_class(k, self, **v)
        self._parent = parent
        self._is_append = True

    def __iter__(self) -> Iterator[T]:
        """
        イテレータ

        :return: iterator
        """
        return self._item.__iter__()

    def items(self):
        """
        dict.items()

        :return: k,v
        """
        return self._item.items()

    def values(self):
        """
        dict.values()

        :return: v
        """
        return self._item.values()

    def keys(self):
        """
        dict.keys()

        :return: k
        """
        return self._item.keys()

    def modify_mark(self, key: str):
        """
        変更されたことをマーク

        :param key: 対象キー名
        :return:
        """
        self._mark.add(key)
        if self._parent:
            self._parent.modify_mark(self.name)

    def flush(self):
        """
        マークをクリア

        :return:
        """
        self._mark.clear()

        # 値が空のときにappendフラグを追ってしまうと新規追加を検知できない
        # そのままupdateされてもmarkが空なのでどのみちクエリは発行されないのでOK
        if not self.is_empty():
            self._is_append = False

    def set_dict(self, d: Dict[str, 'MarkedObjectDict[T]']):
        """
        Dictionaryを直接セットする。このメソッドは更新をマークしない

        :param d: セットするDict
        :return:
        """
        self._item = d

    @classmethod
    def serialize(cls, data: 'MarkedObjectDict[T]'):
        """
        Python → DynamoDB

        :param data: 対象データ
        :return: dict
        """
        ret = {}
        for k, v in data._item.items():
            if not v:
                continue
            ret[k] = v.serialized_value
        return {'M': ret}

    @property
    def serialized_value(self):
        """
        シリアライズされた値

        :return: dict
        """
        return self.__class__.serialize(self)

    @classmethod
    def deserialize(cls, name, target_class: Type[T], data: dict, parent: Markable):
        """
        DynamoDB → Python

        :param name: フィールド名前
        :param target_class: 抱える子タイプ
        :param data: DynamoDBドキュメント
        :param parent: 親Markable
        :return: 生成された値
        """
        ret = cls(name, target_class, parent)
        ret.set_dict({k: target_class.deserialize(k, v, ret) for k, v in data['M'].items()})
        return ret

    def is_empty(self):
        """
        値が空か

        :return: boolean
        """
        for k, v in self._item.items():
            if v and not self._target_class.is_empty(v):
                return False
        return True

    def append(self, value: MarkedObjectWithIndex):
        """
        要素を追加。すでに同一キーの要素が存在した場合は上書きする

        :param value: 追加する値
        :return: None
        """
        self.__setitem__(value.get_hash_value(), value)

    def __getitem__(self, item) -> T:
        """
        値を[xxx]で取得。存在しない場合は例外が発生する。

        :param item: 取得するキー名
        :return: 指定されたキーに対応する要素。
        """
        item_str = str(item)
        return self._item[item_str]

    def __setitem__(self, key, value):
        """
        値を[xxx]で設定。基本的にはappendメソッドを使用すること。

        :param key: セットするキー名
        :param value: セットする値
        :return: None
        """
        key_str = str(key)
        assert (isinstance(value, self._target_class))
        assert (value.get_hash_value() == key_str)
        self._item[key_str] = value
        self.modify_mark(key_str)

    # TODO: 仮対応
    def clear(self):
        self._item: Dict[any, T] = {}
        self._is_append = True

    def create(self, **kwargs) -> T:
        """
        対応する子クラスをインスタンス化して返却する

        :param kwargs: 子クラスのイニシャライザにわたす値
        :return: インスタンス化された子クラス
        """
        return self._target_class(name=None, parent=None, **kwargs)

    def build_update_expression(self, upd: UpdateExpression):
        """
        差分を考慮したupdate条件オブジェクトをセット

        :param upd: セット対象のUpdateExpressionインスタンス
        :return: None
        """

        # 追加されたものは無条件に全て
        if self._is_append:
            upd.set(self.name, self.serialized_value, raw=True)
            return

        if len(self._mark) is 0:
            # 更新されたものは無い
            return

        for update_key in self._mark:
            now_value = self[update_key]
            fc = self._target_class

            n = f'{self.name}.{update_key}'
            if fc.is_empty(now_value):
                # 値が空になった
                upd.remove(n)
                continue
            now_value.build_update_expression(n, upd)



from hatsudenki.packages.command.master.exporter.excel import MasterExcel
from hatsudenki.packages.command.master.tag.loader import MasterTag


class RawYamlRenderer(object):

    def __init__(self, excel_book: MasterExcel):
        self.book = excel_book

    def render(self, out_tag: MasterTag):
        ret = {}
        for sheet in self.book.iter_table_sheet():
            li = []

            for r in sheet.iter():
                # この列内にNoneが含まれてたらエラーにしてやる
                for key, value in r.items():
                    if value is None:
                        raise Exception(
                            f"エクセル:{sheet.book.filename} のシート:{sheet.sheet.title} の{key}に空データが存在します"
                        )

                li.append(r)

            ret[sheet.table.table_name] = li

        return ret



from dataclasses import dataclass

import yaml

from hatsudenki.packages.command.master.table import MasterTable


@dataclass
class MasterDataYamlRenderer:
    master_table: MasterTable

    def render(self):
        return yaml.dump(self.master_table.data, allow_unicode=True, default_flow_style=False)



from pathlib import Path

from hatsudenki.packages.command.hatsudenki.data import HatsudenkiData
from hatsudenki.packages.command.loader.base import LoaderBase


class HatsudenkiLoader(LoaderBase[HatsudenkiData]):
    """
    HatsudenkiDSLローダー
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ext = '.yml'

    def _load(self, path: Path):
        return HatsudenkiData(self.base_path, path, self)

    def setup(self, dir_name=''):
        super().setup(dir_name)



from typing import List

from hatsudenki.packages.table.child import ChildMultiHatsudenkiTable
from hatsudenki.packages.table.define import TableType


class ChildSoloHatsudenkiTable(ChildMultiHatsudenkiTable):

    @classmethod
    def get_table_type(cls):
        return TableType.SingleSoloTable

    @property
    def many_cursor(self):
        raise Exception('Invalid operation.')

    @classmethod
    def resolve_alias(cls, val):
        if val is not None:
            raise Exception('invalid key.')
        # childSoloテーブルは常に固定値
        return cls.get_tag_name()

    @property
    def alias_value(self):
        # childSoloテーブルは常にNone
        return None

    @classmethod
    async def get_iter(cls, hash_val: any, limit=0, prj_exp: List[str] = None):
        raise Exception('invalid operation')

    @classmethod
    def query_parse(cls, query_dict: dict):
        hk = query_dict.get(cls.get_hash_key_name(), None)
        if hk is not None:
            # Hashキーが明示的に指定されている場合はプライマリキーを使用していると仮定する
            ak = query_dict.get(cls.get_alias_key(), None)
            if ak is not None:
                # SoloテーブルはAliasキーを明示的に指定できない
                raise Exception('invalid key.')
            else:
                # SoloテーブルのAliasキーは固定で設定される
                query_dict[cls.get_range_key_name()] = cls.get_tag_name()

        return super().query_parse(query_dict)



import csv
import os
import shutil
import tarfile
from collections import OrderedDict
from os import PathLike
from pathlib import Path
from typing import Iterator, Union

import yaml

# initialize yaml representer
from hatsudenki.packages.command.stdout.output import ToolOutput


def represent_odict(dumper, instance):
    return dumper.represent_mapping('tag:yaml.org,2002:map', instance.items())


yaml.add_representer(OrderedDict, represent_odict)


# initialize yaml constructor
def construct_odict(loader, node):
    return OrderedDict(loader.construct_pairs(node))


yaml.add_constructor('tag:yaml.org,2002:map', construct_odict)


def yaml_load(file_path: Path) -> Union[dict, list]:
    """
    load yaml file
    :param file_path: target file path
    """

    with file_path.open(encoding='utf-8') as file:
        ret = yaml.safe_load(file)

    return ret


def yaml_write(file_path: Path, data: Union[dict, list]):
    write_file(file_path, yaml.dump(data, allow_unicode=True, default_flow_style=False))


def write_file(out_path: Path, data: str, mode: str = 'w'):
    ToolOutput.print(f'write to {out_path}', False)
    out_path.parent.mkdir(parents=True, exist_ok=True)
    with out_path.open(mode=mode, encoding='utf-8') as file:
        file.write(data)


def open_write_stream(file_path: Path, mode: str = 'w'):
    file_path.parent.mkdir(parents=True, exist_ok=True)
    return file_path.open(mode=mode, encoding='utf-8')


def text_load(file_path: Path):
    with file_path.open(mode='r', encoding='utf8') as file:
        text = file.read()
    return text


def write_or_ignore(out_path: PathLike, data: any, mode='w'):
    if os.path.exists(out_path):
        return
    write_file(out_path, data, mode)


def write_csv(out_path: PathLike, data: Iterator[any]):
    p = Path(out_path)
    p.parent.mkdir(parents=True, exist_ok=True)
    with p.open(mode='w', encoding='utf8') as csv_file:
        w = csv.writer(csv_file)
        for l in data:
            w.writerow(l)


def recreate_dir(path: Path):
    shutil.rmtree(path, ignore_errors=True)
    path.mkdir(parents=True, exist_ok=True)


def tar_compress(src: Path, dest: Path):
    dest.parent.mkdir(parents=True, exist_ok=True)
    with tarfile.open(dest, mode='w:gz') as t:
        for f in src.glob('**/*'):
            print(f.relative_to(src))
            t.add(f, str(f.relative_to(src)))


def tar_extract(src: Path, dest: Path):
    dest.parent.mkdir(parents=True, exist_ok=True)
    with tarfile.open(src, mode='r:gz') as t:
        t.extractall(dest)



from dataclasses import dataclass
from pathlib import Path
from typing import List, Dict

from openpyxl.styles import Border, Side, Font, Alignment

from hatsudenki.packages.command.excel.excel_data import ExcelData, CellStyle
from hatsudenki.packages.command.master.loader import MasterTableLoader


class MasterAssignerExcel(ExcelData):

    def __init__(self, file_path: Path, header: List[str]):
        super().__init__(file_path)
        self.header = header
        self.rows = []

    def add_row(self, row: Dict[str, str]):
        self.rows.append(row)

    def render(self, sheet_name: str):
        if sheet_name not in self.book.sheetnames:
            self.book.create_sheet(sheet_name)
        self.write_row(sheet_name, 0, 0, self.header, style=CellStyle.GREEN)

        for idx, r in enumerate(self.rows):
            c = []
            for h in self.header:
                c.append(r.get(h, ''))
            self.write_row(sheet_name, idx + 1, 0, c)


@dataclass
class MasterDataAssignerExcelRenderer:
    master_loader: MasterTableLoader

    def render_excel(self, file_path: Path):
        excel = MasterAssignerExcel(file_path, ['エクセル名', 'シート名', 'テーブル名', 'クライアント', 'サーバー', '企画'])
        excel.create()

        for p, m in self.master_loader.iter():
            excel.add_row({
                'エクセル名': m.excel_name,
                'シート名': m.excel_sheet_name,
                'テーブル名': m.table_name,
                'クライアント': ', '.join(m.assignee_clients),
                'サーバー': ', '.join(m.assignee_servers),
                '企画': ', '.join(m.assignee_planners)
            })
        excel.render('担当者一覧')
        border = Border(
            top=Side(style='thin', color='808080'),
            bottom=Side(style='thin', color='808080'),
            left=Side(style='thin', color='808080'),
            right=Side(style='thin', color='808080')
        )
        font = Font(
            name='MS Gothic',
            size=11
        )
        bold_font = Font(
            name='MS Gothic',
            size=11,
            bold=True
        )

        w = {}

        sheet = excel.get_sheet('担当者一覧')

        for idx, row in enumerate(excel.iter_row('担当者一覧')):
            for c in row:
                c.border = border
                if idx is 0:
                    c.font = bold_font
                    c.alignment = Alignment(horizontal='center')
                else:
                    c.font = font
                v = c.value if c.value else ''

                wi = self._major_width(str(v))
                o = w.get(c.column, 0)
                w2 = max(o, wi)
                if w2 > o:
                    sheet.column_dimensions[c.column].width = w2 + 0.8
                    w[c.column] = w2

        excel.save_file()

    def _major_width(self, text: str):
        from unicodedata import east_asian_width

        width_dict = {
            'F': 2,  # Fullwidth
            'H': 1,  # Halfwidth
            'W': 2,  # Wide
            'Na': 1,  # Narrow
            'A': 2,  # Ambiguous
            'N': 1  # Neutral
        }

        chars = [char for char in text]
        east_asian_width_list = [east_asian_width(char) for char in chars]
        width_list = [width_dict[east_asian_width] for east_asian_width in east_asian_width_list]
        return sum(width_list)



import json

import aiohttp


async def aio_http_post_json(uri: str, body: dict, headers: dict):
    """
    postを投げてjsonでresponse取得

    :param uri: リクエスト先URL
    :param body: リクエストボディ
    :param headers: ヘッダ
    :return: レスポンス
    """
    async with aiohttp.ClientSession() as s:
        async with s.post(uri, data=body, headers=headers) as resp:
            return json.loads(await resp.read())


async def aio_http_put_json(uri: str, body: dict, headers: dict):
    """
    putを投げてjsonでresponse取得

    :param uri: リクエスト先URL
    :param body: リクエストボディ
    :param headers: ヘッダ
    :return: レスポンス
    """
    async with aiohttp.ClientSession() as s:
        async with s.put(uri, data=body, headers=headers) as resp:
            if resp.status != 200:
                raise Exception(await resp.read())
            return json.loads(await resp.read())


async def aio_http_get_json(uri: str, headers: dict):
    """
    getを投げてjsonでresponse取得

    :param uri: リクエスト先URL
    :param headers: ヘッダ
    :return: レスポンス
    """
    async with aiohttp.ClientSession() as s:
        async with s.get(uri, headers=headers) as resp:
            if resp.status != 200:
                raise Exception('status is not 200!!!')
            return json.loads(await resp.read())



import asyncio
import json
import os
import time
from asyncio import sleep
from copy import deepcopy
from datetime import date, datetime
from logging import getLogger
from random import choice
from typing import List, AsyncGenerator, IO

from aioboto3 import Session
from botocore.config import Config

from hatsudenki.packages.counter import QueryCounter
from hatsudenki.packages.expression.base import BaseExpression
from hatsudenki.packages.expression.condition import ConditionExpression, KeyConditionExpression, \
    FilterConditionExpression
from hatsudenki.packages.expression.update import UpdateExpression

_logger = getLogger(__name__)


class HatsudenkiClient(object):
    """
    DynamoDBクライアント
    """

    #: クライアントインスタンス
    _clients = []
    _prefix: str = None
    _is_out_slow_log = False
    _slow_log_duration = 0.1
    _use_profiler = False

    @classmethod
    def dump_count(cls):
        if cls._use_profiler:
            QueryCounter.dump_counter()

    @classmethod
    def setup_by_session(cls, session, prefix: str, endpoint: str = None, connection_num: int = 10):
        con = Config(
            parameter_validation=False,
            max_pool_connections=1,
        )

        print(f'connect dynamodb. endpoint={endpoint} prefix={prefix}, connection_num={connection_num}')
        for _ in range(connection_num):
            cls._clients.append(session.client('dynamodb', endpoint_url=endpoint, config=con, use_ssl=False))
        cls._prefix = prefix

    @classmethod
    def setup(cls, loop, endpoint: str, prefix: str = '', region_name: str = None, aws_access_key_id: str = None,
              aws_secret_access_key: str = None, pool_num: int = 10):
        """
        | 初期設定を行う。
        | すべての処理より先に一度だけ呼び出すこと

        :param loop: 処理を行うIOループ
        :param endpoint: エンドポイント
        :param prefix: テーブルのプリフィックス
        :return: None
        """

        if endpoint:
            # endpointが渡された場合はローカルモードとする
            region = 'local'
            acc_id = 'hogehoge'
            secret = 'fugafuga'
        else:
            region = region_name or os.getenv('AWS_REGION', 'ap-northeast-1')
            acc_id = aws_access_key_id or os.getenv('AWS_ACCESS_KEY_ID')
            secret = aws_secret_access_key or os.getenv('AWS_SECRET_ACCESS_KEY')

        d = {
            'loop': loop
        }

        if region:
            d['region_name'] = region
        if acc_id:
            d['aws_access_key_id'] = acc_id
        if secret:
            d['aws_secret_access_key'] = secret

        ses = Session(
            **d
        )

        cls.setup_by_session(session=ses, prefix=prefix, endpoint=endpoint, connection_num=pool_num)

    @classmethod
    def set_slow_log(cls, flg: bool, duration: float):
        cls._is_out_slow_log = flg
        cls._slow_log_duration = duration

    @classmethod
    def set_use_profiler(cls, flg: bool):
        cls._use_profiler = flg

    @classmethod
    def _take(cls):
        if cls._is_out_slow_log:
            return time.time()
        return None

    @classmethod
    def _cheese(cls, sec, label: str, key: dict):
        if cls._is_out_slow_log:
            p = time.time() - sec
            if p >= cls._slow_log_duration:
                _logger.warning(f'[SLOW] {label} sec={p}', key)
        else:
            return

    @classmethod
    def _get_client(cls):
        return choice(cls._clients)

    @classmethod
    def resolve_table_name(cls, table_name: str, skip: bool = False):
        if skip:
            return table_name
        return cls._prefix + table_name

    @classmethod
    async def create_table(cls, table_name: str, attributes: List[dict], capacity_units: dict, key_schema: List[dict],
                           lsi: List[dict] = None, gsi: List[dict] = None):
        """
        | テーブルを作る
        | DynamoDBはテーブル作成時、キー以外の情報は不要（渡すとエラーになる）

        :param table_name: 対象テーブル名
        :param attributes: 定義する属性
        :param capacity_units: デプロイするキャパシティユニット
        :param key_schema: キー情報
        :param lsi: LocalSecondaryIndex
        :param gsi: GlobalSecondaryIndex
        :return: AWSレスポンス
        """
        ondemand = False
        if capacity_units['ReadCapacityUnits'] == 1 and capacity_units['WriteCapacityUnits'] == 1:
            p = {
                'TableName': cls.resolve_table_name(table_name),
                'AttributeDefinitions': attributes,
                'KeySchema': key_schema,
                'BillingMode': 'PAY_PER_REQUEST'
            }
            ondemand = True
        else:
            p = {
                'TableName': cls.resolve_table_name(table_name),
                'AttributeDefinitions': attributes,
                'KeySchema': key_schema,
                'ProvisionedThroughput': capacity_units
            }

        if lsi is not None and len(lsi) > 0:
            p['LocalSecondaryIndexes'] = lsi

        if gsi is not None and len(gsi) > 0:
            if ondemand:
                gsi = deepcopy(gsi)
                for g in gsi:
                    del g['ProvisionedThroughput']
            p['GlobalSecondaryIndexes'] = gsi

        res = await cls._get_client().create_table(**p)

        return res

    @classmethod
    async def update_table(cls, table_name: str, attributes: List[dict], capacity_units: dict, key_schema: List[dict],
                           lsi: List[dict] = None, gsi: List[dict] = None):
        """
        テーブルを更新する(インデックスとかつくる)
        ※DynamoDBの制限
        - 基本的にGSIの追加・削除およびキャパシティユニットの増減しか行なえません。
        - GSIは追加・削除が行えますが、変更はできません（スループットの変更のみ可能）
        - GSIの操作は同時に1インデックスずつしか行なえません（このメソッドは内部は完了を待ちます）

        :param table_name: 対象テーブル名
        :param attributes: 定義する属性
        :param capacity_units: デプロイするキャパシティユニット
        :param key_schema: キー情報
        :param lsi: LocalSecondaryIndex
        :param gsi: GlobalSecondaryIndex
        :return:
        """

        def _compare_key(old_keys, new_keys):
            old_key_schema = {k['KeyType']: k['AttributeName'] for k in old_keys}
            new_key_schema = {k['KeyType']: k['AttributeName'] for k in new_keys}
            if old_key_schema['HASH'] != new_key_schema['HASH'] \
                    or old_key_schema.get('RANGE', None) != new_key_schema.get('RANGE', None):
                return False
            return True

        def _compare_proj(old_proj, new_proj):
            if old_proj['ProjectionType'] != new_proj['ProjectionType']:
                return False
            if old_proj['ProjectionType'] == 'INCLUDE':
                if set(old_proj['NonKeyAttributes']) != set(new_proj['NonKeyAttributes']):
                    return False
            return True

        # 現在の状態を取得
        desc = await cls.describe_table(table_name)

        # PrimaryKeyは変更できない
        if not _compare_key(desc['KeySchema'], key_schema):
            raise Exception(f'PrimaryIndexが変更されています Table={table_name}')

        # LSIの検証
        now_lsi = {k['IndexName']: k for k in desc.get('LocalSecondaryIndexes', [])}
        new_lsi = {k['IndexName']: k for k in lsi}

        if now_lsi.keys() != new_lsi.keys():
            raise Exception(f'LocalSecondaryIndexはあとから変更できません Table={table_name}')

        for n, i in new_lsi.items():
            li = now_lsi[n]
            if not _compare_key(li['KeySchema'], i['KeySchema']):
                raise Exception(f'LocalSecondaryIndexはあとから変更できません Table={table_name}')
            if not _compare_proj(li['Projection'], i['Projection']):
                raise Exception(f'LocalSecondaryIndexはあとから変更できません Table={table_name}')

        # GSIの差分確認
        now_gsi = {k['IndexName']: k for k in desc.get('GlobalSecondaryIndexes', [])}
        new_gsi = {k['IndexName']: k for k in gsi}

        # 変更の検知
        for k in new_gsi.keys() & now_gsi.keys():
            nk = new_gsi[k]
            ok = now_gsi[k]
            if not _compare_key(ok['KeySchema'], nk['KeySchema']):
                raise Exception(f'GlobalSecondaryIndexはあとから変更できません Table={table_name}')
            if not _compare_proj(ok['Projection'], nk['Projection']):
                raise Exception(f'GlobalSecondaryIndexはあとから変更できません Table={table_name}')

        # 追加の検知
        add_gsi = new_gsi.keys() - now_gsi.keys()
        # 削除の検知
        del_gsi = now_gsi.keys() - new_gsi.keys()

        key_schema_keys = set([k['AttributeName'] for k in key_schema])
        now_attributes = {d['AttributeName']: d for d in desc['AttributeDefinitions']}
        for del_key in del_gsi:
            del_index = now_gsi[del_key]
            del_index_keys = [k['AttributeName'] for k in del_index['KeySchema']]

            atr = [now_attributes[ak] for ak in (key_schema_keys | set(del_index_keys))]

            p = {
                'TableName': cls.resolve_table_name(table_name),
                'AttributeDefinitions': atr,
                'GlobalSecondaryIndexUpdates': [
                    {
                        'Delete': {'IndexName': del_key}
                    }
                ],
            }
            await cls._get_client().update_table(**p)

            # 完了を待つ
            for _ in range(60):
                _desc = await cls.describe_table(table_name)
                _gsi = _desc.get('GlobalSecondaryIndexes', [])
                target = next((g['IndexStatus'] for g in _gsi if g['IndexName'] == del_key), None)
                if target is None:
                    break
                await sleep(1)

        # GSIの作成は一回ずつしかできない
        is_ondemand = desc.get('BillingModeSummary') and desc.get('BillingModeSummary')[
            'BillingMode'] == 'PAY_PER_REQUEST'

        work_attributes = {k['AttributeName']: k for k in attributes}
        for add_key in add_gsi:
            add_index = new_gsi[add_key]
            if is_ondemand and add_index.get('ProvisionedThroughput'):
                # ondemand時はCUの設定は不要
                del add_index['ProvisionedThroughput']

            add_index_keys = [k['AttributeName'] for k in add_index['KeySchema']]

            atr = [work_attributes[ak] for ak in (key_schema_keys | set(add_index_keys))]

            p = {
                'TableName': cls.resolve_table_name(table_name),
                'AttributeDefinitions': atr,
                'GlobalSecondaryIndexUpdates': [
                    {
                        'Create': add_index
                    }
                ],
            }

            await cls._get_client().update_table(**p)

            # 完了を待つ
            for _ in range(60):
                _desc = await cls.describe_table(table_name)
                _gsi = _desc['GlobalSecondaryIndexes']
                target = next((g['IndexStatus'] for g in _gsi if g['IndexName'] == add_key), None)
                if target is None:
                    raise Exception('Invalid GSI!!!')
                if target == 'ACTIVE':
                    break
                await sleep(1)

    @classmethod
    async def drop_table(cls, table_name: str):
        _start = cls._take()
        p = {
            'TableName': cls.resolve_table_name(table_name)
        }
        res = await cls._get_client().delete_table(**p)
        cls._cheese(_start, f'drop_table {table_name}', p)
        return res

    @classmethod
    async def describe_table(cls, table_name: str):
        """
        テーブル情報を取得

        :param table_name: 対象テーブル名
        :return: テーブル情報を格納した辞書配列。AWSのレスポンス参照
        """

        _start = cls._take()
        p = {
            'TableName': cls.resolve_table_name(table_name)
        }
        res = await cls._get_client().describe_table(**p)
        cls._cheese(_start, f'describe_table {table_name}', p)
        return res['Table']

    @classmethod
    async def list_tables(cls, start_table_name: str = None) -> List[str]:
        """
        テーブル一覧を取得

        :return: テーブル名のリスト
        """
        _start = cls._take()
        st = cls._prefix or ''

        if start_table_name:
            st += start_table_name

        p = {
            **({'ExclusiveStartTableName': st} if len(st) is not 0 else {})
        }

        res = await cls._get_client().list_tables(**p)
        cls._cheese(_start, f'list_tables {start_table_name}', p)
        return [i.replace(st, '') for i in res.get('TableNames')]

    @classmethod
    async def get_item(cls, table_name: str, key: dict, prj: List[str] = None):
        """
        キー指定式を使用して一件取得

        :param table_name: テーブル名
        :param key: キー指定式辞書配列
        :return: アイテム情報を格納した辞書配列。AWSレスポンス参照。アイテムが存在しない場合はNone
        """

        _start = cls._take()
        p = {
            'TableName': cls.resolve_table_name(table_name),
            'Key': key
        }

        if prj is not None:
            p['ProjectionExpression'] = ','.join(prj)

        res = await cls._get_client().get_item(**p)

        if cls._use_profiler:
            p['ReturnConsumedCapacity'] = 'INDEXES'

        r = res.get('Item', None)

        cls._cheese(_start, f'get_item, {table_name}', p)
        if cls._use_profiler:
            if key.get('kind', None):
                QueryCounter.count('get', f"{table_name} - {key['kind']['S'].split('//')[0]}")
            else:
                QueryCounter.count('get', f"{table_name}")

            consumed_cu = res.get('ConsumedCapacity', False)
            if consumed_cu:
                QueryCounter.count_read_ccu(consumed_cu)

        return r

    @classmethod
    async def put_item(cls, table_name: str, item: dict, condition: ConditionExpression = None, raw_table_name=False):
        """
        | アイテムを作成
        | 更新条件式が成立しなかった場合は例外が発生する

        :param table_name: 対象テーブル名
        :param item: シリアライズされたアイテム情報を格納した辞書配列
        :param condition: 更新条件式インスタンス
        :param raw_table_name: テーブル名にプリフィックスを付与しない
        :return: AWSレスポンス
        """
        _start = cls._take()

        p = {
            'TableName': cls.resolve_table_name(table_name, raw_table_name),
            'Item': item,

            **(condition.to_parameter() if condition is not None else {})
        }
        if cls._use_profiler:
            p['ReturnConsumedCapacity'] = 'INDEXES'

        res = await cls._get_client().put_item(**p)

        cls._cheese(_start, f'put_item {table_name}', p)
        if cls._use_profiler:
            if item.get('kind', None):
                QueryCounter.count('put', f"{table_name} - {item['kind']['S'].split('//')[0]}")
            else:
                QueryCounter.count('put', f"{table_name}")

            consumed_cu = res.get('ConsumedCapacity', False)
            if consumed_cu:
                QueryCounter.count_write_ccu(consumed_cu)

        return res

    @classmethod
    async def delete_item(cls, table_name, key: dict, condition: ConditionExpression = None, raw_table_name=False):
        _start = cls._take()

        p = {
            'TableName': cls.resolve_table_name(table_name, raw_table_name),
            'Key': key,
            **(condition.to_parameter() if condition is not None else {})
        }

        res = await cls._get_client().delete_item(**p)

        cls._cheese(_start, f'delete_item {table_name}', p)
        return res

    @classmethod
    async def update_item(cls, table_name: str, key: dict, update: UpdateExpression,
                          condition: ConditionExpression = None, raw_table_name=False):
        """
        アイテムを更新

        :param table_name: 対象テーブル名
        :param key: ハッシュキー及びレンジキーの指定式
        :param update: 更新値式インスタンス
        :param condition: 更新条件式インスタンス
        :return: AWSレスポンス
        """
        _start = cls._take()

        p = {
            'TableName': cls.resolve_table_name(table_name, raw_table_name),
            'Key': key,
            **(update.to_parameter(condition))
        }

        if cls._use_profiler:
            p['ReturnConsumedCapacity'] = 'INDEXES'
        res = await cls._get_client().update_item(**p)

        cls._cheese(_start, f'update_item {table_name}', p)
        if cls._use_profiler:
            if key.get('kind', None):
                QueryCounter.count('update', f"{table_name} - {key['kind']['S'].split('//')[0]}")
            else:
                QueryCounter.count('update', f"{table_name}")
            consumed_cu = res.get('ConsumedCapacity', False)
            if consumed_cu:
                QueryCounter.count_write_ccu(consumed_cu)
        return res

    @classmethod
    async def query(cls, table_name: str, key_cond: KeyConditionExpression, filter_cond: BaseExpression = None,
                    use_index_name: str = None, limit=0, prj: List[str] = None, raw_table_name=False, max_retry=100):
        """
        条件式を複数指定する検索。複数件のアイテムを返却する

        :param table_name: 対象テーブル名
        :param key_cond: キー絞り込み
        :param filter_cond: フィルター
        :param use_index_name: 使用するインデックス名
        :param limit: リミット
        :param prj: プロジェクション情報
        :param raw_table_name: テーブル名にプリフィックスを付与しない
        :return: AWSレスポンス
        """
        ret = []
        last_key = None

        for query_num in range(max_retry):
            _start = cls._take()

            p = {
                'TableName': cls.resolve_table_name(table_name, raw_table_name),
                **({'IndexName': use_index_name} if use_index_name is not None else {}),
                **({'Limit': limit} if limit > 0 is not None else {}),
                **(BaseExpression.merge(key_cond, filter_cond))
            }
            if prj is not None:
                p['ProjectionExpression'] = ','.join(prj)

            if cls._use_profiler:
                p['ReturnConsumedCapacity'] = 'INDEXES'

            res = await cls._get_client().query(**p)

            cls._cheese(_start, f'query {table_name}', p)

            if cls._use_profiler:
                if p['ExpressionAttributeValues'].get(':key_value__1', None):
                    QueryCounter.count('query',
                                       f"{table_name} - {p.get('IndexName')} - {p['ExpressionAttributeValues'][':key_value__1']}")
                else:
                    QueryCounter.count('query', f"{table_name}")
                consumed_cu = res.get('ConsumedCapacity', False)
                if consumed_cu:
                    QueryCounter.count_read_ccu(consumed_cu)
            ret.extend(res['Items'])
            last_key = res.get('LastEvaluatedKey')
            if last_key is None:
                break
            else:
                _logger.warning(f'over 1MB response. query next {query_num}')
                await asyncio.sleep(0.1)

        return ret

    @classmethod
    async def scan(cls, table_name: str, filter_cond: FilterConditionExpression = None, limit=20,
                   prj: List[str] = None, raw_table_name=False):
        """
        フルスキャンを行う。危険なので使用には気をつけること。

        :param table_name: 対象テーブル名
        :param filter_cond: 絞り込み条件式
        :return: AWSレスポンス
        """
        _start = cls._take()
        p = {
            'TableName': cls.resolve_table_name(table_name, raw_table_name),
            **(filter_cond.to_parameter() if filter_cond is not None else {}),
            'Limit': limit
        }
        if prj is not None:
            p['ProjectionExpression'] = ','.join(prj)

        res = await cls._get_client().scan(**p)
        ret = res['Items']
        cls._cheese(_start, f'scan {table_name}', p)
        if len(ret) >= limit:
            return ret[:limit]

        while 'LastEvaluatedKey' in res:
            _start = cls._take()
            p['ExclusiveStartKey'] = res['LastEvaluatedKey']
            res = await cls._get_client().scan(**p)
            ret.extend(res['Items'])
            cls._cheese(_start, f'scan {table_name}', p)
            if len(ret) >= limit:
                return ret[:limit]

        return ret

    @classmethod
    async def scan_generator(cls, table_name: str, filter_cond: FilterConditionExpression = None, tick=100,
                             prj: List[str] = None, raw_table_name=False) -> AsyncGenerator[List[dict], None]:

        p = {
            'TableName': cls.resolve_table_name(table_name, raw_table_name),
            **(filter_cond.to_parameter() if filter_cond is not None else {}),
            'Limit': tick
        }
        if prj is not None:
            p['ProjectionExpression'] = ','.join(prj)

        res = await cls._get_client().scan(**p)
        ret = res['Items']
        yield ret

        if 'LastEvaluatedKey' not in res:
            return

        while 'LastEvaluatedKey' in res:
            p['ExclusiveStartKey'] = res['LastEvaluatedKey']
            res = await cls._get_client().scan(**p)
            yield res['Items']

    @classmethod
    async def set_ttl_mode(cls, table_name: str, attr_name: str, set_mode=True, raw_table_name=False):
        _start = cls._take()
        p = {
            'TableName': cls.resolve_table_name(table_name),
            'TimeToLiveSpecification': {
                'Enabled': set_mode,
                'AttributeName': attr_name
            }
        }
        res = await cls._get_client().update_time_to_live(**p)

        cls._cheese(_start, f'set_ttl_mode {table_name}', p)
        return res

    @classmethod
    async def batch_get_item(cls, request_items: dict, max_retry=100):
        ret = {}
        req_items = request_items

        for cnt in range(max_retry):
            _start = cls._take()
            p = {
                'RequestItems': req_items
            }
            if cls._use_profiler:
                p['ReturnConsumedCapacity'] = 'INDEXES'

            res = await cls._get_client().batch_get_item(**p)
            cls._cheese(_start, 'batch_get_item', p)

            if cls._use_profiler:
                QueryCounter.count('batch_get')
                consumed_cu_list = res.get('ConsumedCapacity', False)
                for consumed_cu in consumed_cu_list:
                    QueryCounter.count_read_ccu(consumed_cu)

            # 値の回収
            for k, i in res['Responses'].items():
                if k in res:
                    ret[k].extend(i)
                else:
                    ret[k] = i

            # unprocessが存在した場合はもう一回
            req_item = res['UnprocessedKeys']
            if not req_item:
                break
            else:
                _logger.warning(f'exists UnprocessedKeys. retrying...{cnt}')
        return ret

    @classmethod
    async def batch_write_item(cls, request_items: dict, max_retry=100):
        req_items = request_items
        for cnt in range(max_retry):
            _start = cls._take()
            p = {
                'RequestItems': req_items
            }

            if cls._use_profiler:
                p['ReturnConsumedCapacity'] = 'INDEXES'

            res = await cls._get_client().batch_write_item(**p)
            cls._cheese(_start, 'batch_write_item', p)

            if cls._use_profiler:
                QueryCounter.count('batch_write')
                consumed_cu_list = res.get('ConsumedCapacity', False)
                for consumed_cu in consumed_cu_list:
                    QueryCounter.count_write_ccu(consumed_cu)

            # unprocessが存在した場合はもう一回
            req_items = res['UnprocessedItems']
            if not req_items:
                break
            else:
                _logger.warning(f'UnprocessedItems foud. next={cnt}')
                await asyncio.sleep(0.1)

    @classmethod
    async def transaction_write(cls, items: list):
        _start = cls._take()
        p = {
            'TransactItems': items
        }

        if cls._use_profiler:
            p['ReturnConsumedCapacity'] = 'INDEXES'
        res = await cls._get_client().transact_write_items(**p)
        cls._cheese(_start, 'transact_write', p)

        if cls._use_profiler:
            QueryCounter.count('transact_write')
            consumed_cu_list = res.get('ConsumedCapacity', False)
            for consumed_cu in consumed_cu_list:
                QueryCounter.count_write_ccu(consumed_cu)
        return res

    @classmethod
    async def transaction_get(cls, items: list) -> List[dict]:

        _start = cls._take()
        p = {
            'TransactItems': items
        }

        if cls._use_profiler:
            p['ReturnConsumedCapacity'] = 'INDEXES'

        res = await cls._get_client().transact_get_items(**p)
        cls._cheese(_start, 'transact_get', p)
        if cls._use_profiler:
            QueryCounter.count('transact_get')
            consumed_cu_list = res.get('ConsumedCapacity', False)
            for consumed_cu in consumed_cu_list:
                QueryCounter.count_write_ccu(consumed_cu)
        return res['Responses']

    def __init__(self):
        # インスタンス化する必要はない
        raise Exception('TatsumakiClient is mono state.')

    @classmethod
    async def die(cls):
        t = [c.close() for c in cls._clients]
        await asyncio.gather(*t)

    @classmethod
    async def export_json(cls, table_name: str, out_stream: IO, tick=100, raw_table_name=False):
        def json_serial(obj):
            if isinstance(obj, (datetime, date)):
                return obj.isoformat()
            raise TypeError("Type %s not serializable" % type(obj))

        desc = await cls.describe_table(table_name)
        out_stream.write(json.dumps(desc, ensure_ascii=False, default=json_serial) + '\n')

        gen = cls.scan_generator(table_name=table_name, filter_cond=None, tick=tick, prj=None,
                                 raw_table_name=raw_table_name)
        async for l in gen:
            for i in l:
                out_stream.write(json.dumps(i, ensure_ascii=False) + '\n')

    @classmethod
    async def import_json(cls, table_name: str, in_stream: IO, recreate=False):
        desc = json.loads(in_stream.readline())
        if recreate:
            p = {
                'TableName': cls.resolve_table_name(table_name),
                'AttributeDefinitions': desc['AttributeDefinitions'],
                'KeySchema': desc['KeySchema']
            }

            if desc['ProvisionedThroughput']:
                p['ProvisionedThroughput'] = {
                    'ReadCapacityUnits': desc['ProvisionedThroughput']['ReadCapacityUnits'],
                    'WriteCapacityUnits': desc['ProvisionedThroughput']['WriteCapacityUnits']
                }
            else:
                p['BillingMode'] = 'PAY_PER_REQUEST'

            if 'GlobalSecondaryIndexes' in desc:
                gsi = []
                for g in desc['GlobalSecondaryIndexes']:
                    c = {
                        'IndexName': g['IndexName'],
                        'KeySchema': g['KeySchema'],
                        'Projection': g['Projection']
                    }
                    if 'ProvisionedThroughput' in g:
                        c['ProvisionedThroughput'] = {
                            'ReadCapacityUnits': g['ProvisionedThroughput']['ReadCapacityUnits'],
                            'WriteCapacityUnits': g['ProvisionedThroughput']['WriteCapacityUnits']
                        }
                    gsi.append(c)
                p['GlobalSecondaryIndexes'] = gsi

            if 'LocalSecondaryIndexes' in desc:
                lsi = []
                for l in desc['LocalSecondaryIndexes']:
                    c = {
                        'IndexName': l['IndexName'],
                        'KeySchema': l['KeySchema'],
                        'Projection': l['Projection']
                    }
                    lsi.append(c)
                p['LocalSecondaryIndexes'] = lsi

            # 作り直す
            await cls.drop_table(table_name)
            await cls._get_client().create_table(**p)

        idx = 0
        items = []
        for line in in_stream:
            d = json.loads(line)
            items.append({'PutRequest': {'Item': d}})
            idx += 1
            if idx >= 25:
                await cls.batch_write_item(
                    request_items={cls.resolve_table_name(table_name): items}
                )
                items = []

        if len(items) > 0:
            await cls.batch_write_item(
                request_items={cls.resolve_table_name(table_name): items}
            )



from datetime import datetime
from typing import Generic, TypeVar

T = TypeVar('T')


class BaseMasterColumn(Generic[T]):
    def __init__(self, name: str):
        self.name = name

    def convert(self, value: any) -> T:
        return value


class MasterColumnInt(BaseMasterColumn[int]):

    def convert(self, value: any) -> T:
        return int(value)


class MasterColumnString(BaseMasterColumn[str]):
    def convert(self, value: any) -> T:
        return str(value)


class MasterColumnEnum(BaseMasterColumn[int]):
    def convert(self, value: any) -> T:
        return int(value)


class MasterColumnRelation(BaseMasterColumn[str]):

    def __init__(self, name: str, to: str):
        super().__init__(name)
        self.to = to


class MasterColumnDate(BaseMasterColumn[datetime]):

    def convert(self, value: any) -> T:
        return datetime.fromtimestamp(value)


class MasterColumnSelect(BaseMasterColumn[int]):
    pass


class MasterColumnChose(BaseMasterColumn[str]):
    pass



import datetime
from dataclasses import dataclass, field
from logging import getLogger
from typing import Type, Dict, List

from botocore.exceptions import ClientError

from hatsudenki.packages.client import HatsudenkiClient
from hatsudenki.packages.table.index import PrimaryIndex, GSI, LSI, IndexBase
from hatsudenki.packages.table.solo import SoloHatsudenkiTable

_logger = getLogger(__name__)


@dataclass
class CollectionIndex:
    primary: PrimaryIndex
    gsi: List[GSI] = field(default_factory=list)
    lsi: List[LSI] = field(default_factory=list)


class DateResolver(object):
    def get_now(self):
        raise NotImplementedError()


class DefaultDateResolver(DateResolver):
    def get_now(self):
        return datetime.datetime.now()


class TableManager(object):
    _all_collections: Dict[str, List[SoloHatsudenkiTable]] = {}
    _all_tables: Dict[str, SoloHatsudenkiTable] = {}
    _indexes: Dict[str, CollectionIndex] = {}

    @classmethod
    def register(cls, tbl: Type[SoloHatsudenkiTable]):
        tbl_name = tbl.Meta.table_name
        if tbl_name is None or len(tbl_name) is 0:
            return False

        if tbl_name in cls._all_tables:
            return False
        cls._all_tables[tbl.get_table_name()] = tbl
        c = cls._all_collections.get(tbl.Meta.collection_name, [])
        c.append(tbl)
        cls._all_collections[tbl.Meta.collection_name] = c

        i = cls._indexes.get(tbl.Meta.collection_name, None)
        if i is None:
            i = CollectionIndex(tbl.Meta.primary_index, [], [])
            cls._indexes[tbl.Meta.collection_name] = i

        for v in tbl.LSIndex.__dict__.values():
            if type(v) is LSI:
                i.lsi.append(v)
        for v in tbl.GSIndex.__dict__.values():
            if type(v) is GSI:
                i.gsi.append(v)
        return True

    @classmethod
    def get_by_table_name(cls, table_name: str):
        return cls._all_tables[table_name]

    @classmethod
    def get_index(cls, collection_name: str):
        return cls._indexes[collection_name]

    @classmethod
    def find_field_class(cls, collectioln_name: str, key_name: str):
        i = cls._all_collections[collectioln_name]
        for t in i:
            f = t.get_field_class(key_name)
            if f:
                return f
        return None

    @classmethod
    def _to_key_schema(cls, tbl: SoloHatsudenkiTable, idx: IndexBase):
        i = cls._all_collections[tbl.Meta.collection_name]
        ks = []
        h = cls.find_field_class(tbl.get_collection_name(), idx.hash_key)
        ks.append(h.key_schema('HASH'))
        if idx.range_key is not None:
            print(f'{tbl.get_collection_name()}-{tbl.get_table_name()} {idx.range_key}')
            r = cls.find_field_class(tbl.get_collection_name(), idx.range_key)
            ks.append(r.key_schema('RANGE'))

        return ks

    @classmethod
    async def drop_table(cls, collection_name: str):
        await HatsudenkiClient.drop_table(collection_name)

    @classmethod
    async def all_drop(cls):
        for name, tbls in cls._all_collections.items():
            for tbl in tbls:
                if not tbl.Meta.is_root:
                    continue
                try:
                    await cls.drop_table(tbl.get_collection_name())
                except ClientError as e:
                    # テーブル無いエラーは無視して構わない
                    if e.response['Error']['Code'] != 'ResourceNotFoundException':
                        raise e

    @classmethod
    async def create_tables(cls, drop=False, labels=['default']):
        _logger.info('begin create table.')
        for name, tbls in cls._all_collections.items():
            for tbl in tbls:
                if tbl.Meta.label not in labels:
                    continue
                if not tbl.Meta.is_root:
                    continue
                _logger.info(f'create {name}')
                print(f'create {name}')

                attrs = set()
                idx = cls._indexes[tbl.Meta.collection_name]
                key_schema = cls._to_key_schema(tbl, idx.primary)
                attrs |= idx.primary.use_keys

                lsi = []
                for l in idx.lsi:
                    s = cls._to_key_schema(tbl, l)
                    lsi.append({
                        'IndexName': l.name,
                        'KeySchema': s,
                        'Projection': l.to_projection_schema()
                    })

                    attrs |= l.use_keys
                gsi = []
                for g in idx.gsi:
                    # GSIの生成
                    s = cls._to_key_schema(tbl, g)
                    gsi.append({
                        'IndexName': g.name,
                        'KeySchema': s,
                        'ProvisionedThroughput': g.capacity_units.to_schema(),
                        'Projection': g.to_projection_schema()
                    })
                    attrs |= g.use_keys

                # 使用キーは定義リストに入っている必要がある
                attr_def = [cls.find_field_class(tbl.get_collection_name(), key).attribute_define for key in attrs]

                if drop:
                    _logger.warning(f'drop table {name}')
                    # 作る前にドロップする
                    try:
                        await cls.drop_table(tbl.get_collection_name())
                    except ClientError as e:
                        # テーブル無いエラーは無視して構わない
                        if e.response['Error']['Code'] != 'ResourceNotFoundException':
                            raise e
                try:
                    # 作る
                    res = await HatsudenkiClient.create_table(
                        tbl.get_collection_name(), attr_def, idx.primary.capacity_units.to_schema(), key_schema, lsi,
                        gsi)
                except ClientError as e:
                    # テーブルすでにあるエラーは無視して構わない
                    if e.response['Error']['Code'] != 'ResourceInUseException':
                        raise e

                    # Index情報を合わせる
                    await HatsudenkiClient.update_table(
                        tbl.get_collection_name(), attr_def,
                        idx.primary.capacity_units.to_schema(), key_schema, lsi,
                        gsi
                    )

                    res = None

                # TTLが設定されているか
                ttl_key = next((k for k, v in tbl._attributes.items() if v.is_ttl), None)
                if ttl_key is not None:
                    try:
                        await HatsudenkiClient.set_ttl_mode(tbl.get_collection_name(), ttl_key, True)
                    except:
                        pass

    @classmethod
    def get_table_num(cls):
        return len(cls._all_tables.keys())

    @classmethod
    def get_collection_num(cls):
        return len(cls._all_collections.keys())



import enum
from dataclasses import dataclass
from typing import List


class IndexType(enum.Enum):
    """
    Indexの種類
    """
    Primary = enum.auto()
    Local = enum.auto()
    Global = enum.auto()


@dataclass
class CapacityUnits:
    """
    CapacityUnitの詳細
    """
    read: int = 1
    write: int = 1

    def to_schema(self):
        """
        query用のスキーマを取得

        :return: dict
        """
        return {
            'ReadCapacityUnits': self.read,
            'WriteCapacityUnits': self.write
        }


class IndexBase(object):
    """
    Indexのベース
    """

    def __init__(self, name: str, hash_key: str, range_key: str = None):
        """
        init

        :param name: インデックス名
        :param hash_key: ハッシュキー名
        :param range_key: レンジキー名
        """
        self.name = name
        self.hash_key = hash_key
        self.range_key = range_key
        self.use_keys = {hash_key}
        if range_key is not None:
            self.use_keys.add(range_key)

    def check(self, hash_key: str, range_key: str = None):
        """
        使用できるかのチェック

        :param hash_key: ハッシュキー名
        :param range_key: レンジキー名
        :return: bool
        """

        # HASHのみで検索も可能なので、RANGEがNONEでも許容する
        if range_key is None:
            return self.hash_key == hash_key
        return self.hash_key == hash_key and self.range_key == range_key


class PrimaryIndex(IndexBase):
    """
    プライマリインデックス
    """

    def __init__(self, read_cap: int, write_cap: int, **kwargs):
        """
        init

        :param read_cap: リードキャパシティ
        :param write_cap: ライトキャパシティ
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.capacity_units = CapacityUnits(read=read_cap, write=write_cap)


class SecondaryIndex(IndexBase):
    """
    セカンダリインデックス
    """

    def __init__(self, projection_keys: List[str], **kwargs):
        """
        init

        :param projection_keys: プロジェクションするキーリスト
        :param kwargs:
        """

        super().__init__(**kwargs)
        self.projection = projection_keys

    def to_projection_schema(self):
        """
        プロジェクションスキーマに変換

        :return: dict
        """
        if len(self.projection) > 0:
            return {
                'ProjectionType': 'INCLUDE',
                'NonKeyAttributes': self.projection
            }
        else:
            # TODO: ひとまず省略した場合はALLとする
            return {
                'ProjectionType': 'ALL',
            }


class GSI(SecondaryIndex):
    """
    グローバルセカンダリインデックス
    """

    def __init__(self, read_cap: int, write_cap: int, **kwargs):
        super().__init__(**kwargs)
        self.capacity_units = CapacityUnits(read=read_cap, write=write_cap)


class LSI(SecondaryIndex):
    """
    ローカルセカンダリインデックス
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



import struct

import msgpack
from lz4 import block

from hatsudenki.packages.command.master.exporter.pack.yaml import MasterExportYaml


class PackRenderer(object):
    def __init__(self, yaml_data: MasterExportYaml):
        self.data = yaml_data

    def render(self):
        cpy = []
        key_order = [c.column_name for c in self.data.table.columns.values() if not c.is_no_pack]
        key_order.extend([c.column_name for c in self.data.table.shadow_columns.values() if not c.is_no_pack])

        for row_idx, row in enumerate(self.data.data):
            c = []
            for key in key_order:
                c.append(row[key])
            cpy.append(c)

        pack = msgpack.packb([cpy])
        pack_len = len(pack)
        # ブロックフォーマットで圧縮。MsgPackにサイズを付与するため、compressバイナリには含めない
        comp = block.compress(pack, mode='default', store_size=False)

        # ヘッダ作成
        # Pythonでは明示的にExt32が指定できない？ので自前で突っ込む
        t = struct.pack('B', 0xc9)
        # サイズ情報がタイプ＋4バイト付与されるため、5バイト膨らむ
        t += struct.pack('>I', len(comp) + 5)
        # msgpack-csharpは99番をLZ4コンテナとして使っているようだ
        t += struct.pack('B', 99)
        # 解凍後サイズを頭に突っ込む
        t += struct.pack('B', 0xd2)
        t += struct.pack('>I', pack_len)
        # ヘッダとボディを結合
        t += comp

        return t

    def yaml_render(self):
        ret = []
        for row_idx, row in enumerate(self.data.data):
            ret.append(row)
        return ret



class BaseCommand(object):
    def __init__(self, stdout=None, stderr=None, no_color=False):
        pass

    async def handle(self, *args, **options):
        print('hoge')

    def add_arguments(self, parser):
        pass



from typing import Type, TypeVar

from hatsudenki.packages.client import HatsudenkiClient
from hatsudenki.packages.expression.condition import ConditionExpression
from hatsudenki.packages.expression.update import UpdateExpression
from hatsudenki.packages.table.multi import MultiHatsudenkiTable
from hatsudenki.packages.table.solo import SoloHatsudenkiTable

Table = TypeVar('Table', SoloHatsudenkiTable, MultiHatsudenkiTable)


class DirectUpdate(object):

    def __init__(self, model: Type[SoloHatsudenkiTable]):
        self.model = model
        self.exp = UpdateExpression()
        self.cond = ConditionExpression()
        self.key = None

    def set_key_raw(self, key):
        self.key = key

    def set_key(self, hash_val, range_val):
        self.key = self.model.get_serialized_key(hash_val, range_val)

    def set_key_direct(self, **vals):
        self.key = {}

        ks = self.model.get_primary_key_names()
        for k, v in vals.items():
            if k in ks:
                self.key[k] = self.model.get_field_class(k).serialize(v)
            else:
                Exception(f'{k} is not hash_key or range_key')

    async def exec(self):
        await HatsudenkiClient.update_item(
            self.model.get_collection_name(),
            self.key,
            self.exp,
            self.cond
        )

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.exec()



from hatsudenki.packages.command.hatsudenki.field import HatsudenkiFieldBase

_DYNAMO_FIELDS = {}


class FieldFactory(object):
    _fields = {}

    @classmethod
    def create_field(cls, name: str, data: dict, parent) -> HatsudenkiFieldBase:
        return cls._fields[data['type']](name, data, parent)

    @classmethod
    def register(cls, field: HatsudenkiFieldBase):
        cls._fields[field.TypeStr] = field


def field(key, val_type):
    def _wrap(cls):
        _DYNAMO_FIELDS[key] = cls
        cls.TypeStr = key
        cls.PythonValueTypeStr = val_type

        FieldFactory.register(cls)
        return cls

    return _wrap



from __future__ import annotations

from typing import Generic, TypeVar

from hatsudenki.packages.command.stdout.output import IndentString

T = TypeVar('T')


class FileRenderer(object):
    def __init__(self):
        self.units = []

    def render_header(self) -> IndentString:
        return IndentString()

    def add_unit(self, *unit: RenderUnit):
        [self.units.append(u) for u in unit]

    def render_units(self):
        u = IndentString()
        [u.add(un.render()) for un in self.units]
        return u

    def render(self):
        body = IndentString()
        body.add(self.render_header())
        body.add(self.render_units())

        return body.render()


class RenderUnit(Generic[T]):
    def __init__(self, data: T):
        self.data = data

    def render(self) -> IndentString:
        return IndentString()



from typing import Generic, Type, Union, TypeVar

from hatsudenki.packages.expression.update import UpdateExpression
from hatsudenki.packages.marked import Markable

T = TypeVar('T')


class MarkedObjectMap(Generic[T]):
    def __init__(self, name: str, target_class: Type[T], parent: Markable, item: Union[T, dict, None] = None):
        """
        イニシャライザ

        :param name: 名前
        :param target_class: 対象クラス名
        :param parent: 親オブジェクト
        :param item: 初期化パラメータ
        """
        self.name = name
        self._target_class: Type[T] = target_class
        self._mark = set()
        self._is_assign = False
        self._parent = parent
        self._item: T = None
        if isinstance(item, target_class):
            # 値が直接渡された
            self._item = item
        elif type(item) is dict:
            # dict
            self._item = target_class(name, self, **item)

    def assign(self, value: T):
        """
        値をアサインする

        :param value: アサインする値
        :return: None
        """
        self._item = value
        # 名前をセットしておく
        value._name = self.name
        # マーク
        self.assign_mark()

    @property
    def value(self) -> T:
        """
        アサインされている値

        :return: T
        """

        return self._item

    def assign_mark(self):
        """
        値をアサインされたことをマークする

        :return: None
        """
        self._is_assign = True
        if self._parent:
            self._parent.modify_mark(self.name)

    def modify_mark(self, key):
        """
        値が更新されたことをマークする
        :param key:
        :return:
        """
        self._mark.add(key)
        if self._parent:
            self._parent.modify_mark(self.name)

    def flush(self):
        """
        変更情報をクリア

        :return: None
        """
        self._mark.clear()
        self._is_assign = False

    @classmethod
    def serialize(cls, data: 'MarkedObjectMap[T]'):
        """
        DynamoDBへシリアライズ

        :param data: 対象データ
        :return: DynamoDB dict
        """

        return data._item.serialized_value

    @property
    def serialized_value(self):
        """
        DynamoDB dictにシリアライズされた値

        :return: DynamoDB dict
        """
        return self.__class__.serialize(self)

    @classmethod
    def deserialize(cls, name, target_class: Type[T], data: dict, parent: Markable):
        """
        pythonにデシリアライズ
        :param name: 名前
        :param target_class: 格納対象クラスタイプ
        :param data: dynamoDB dict
        :param parent: 親オブジェクト
        :return: 自分自身をインスタンス化したものy
        """
        ret = cls(name, target_class, parent)
        ret._item = target_class.deserialize(name, data, ret)
        return ret

    def build_update_expression(self, upd: UpdateExpression):
        """
        更新箇所を考慮したクエリをビルド。
        追加と更新が同時に行われた場合は例外が発生する

        :param upd: UpdateExpressionインスタンス
        :return: None
        """
        if self._item is None:
            return

        if self._is_assign:
            upd.set(self.name, self.serialized_value, raw=True)
            return

        # 更新キーを考慮して小要素を直接セットするクエリをビルド
        self._item.build_update_expression(self.name, upd)

    def create(self, **kwargs) -> T:
        """
        格納可能なインスタンスを生成して返す（追加はしない）

        :param kwargs: 初期値
        :return: 生成されたインスタンス
        """
        return self._target_class(name=None, parent=None, **kwargs)

    def is_empty(self):
        return self._target_class.is_empty(self._item)



import asyncio
import itertools
import pprint
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from pprint import pprint
from typing import List, Dict, Type, TypeVar

from hatsudenki.packages.client import HatsudenkiClient
from hatsudenki.packages.table.define import TableType
from hatsudenki.packages.table.solo import SoloHatsudenkiTable


@dataclass
class BatchGetTask:
    table_name: str = None
    keys: dict = None


class BatchWriteKind(Enum):
    Put = 'PutRequest'
    Delete = 'DeleteRequest'


@dataclass
class BatchWriteTask:
    table_name: str = None
    item: dict = None
    kind: BatchWriteKind = None

    @property
    def query(self):
        return {
            self.kind.value: {'Item': self.item}
        } if self.kind is BatchWriteKind.Put else {
            self.kind.value: {'Key': self.item}
        }

    @property
    def real_table_name(self):
        return HatsudenkiClient.resolve_table_name(self.table_name)


class QueryBatchGetItem(object):
    def __init__(self):
        self._task: List[BatchGetTask] = []

    def append(self, table: SoloHatsudenkiTable):
        self._task.append(
            BatchGetTask(table_name=table._collection_name, keys=table.serialized_key))

    def exec_query(self, limit=100):
        query = defaultdict(lambda: {'Keys': []})
        w = 0
        for task in self._task:
            query[HatsudenkiClient.resolve_table_name(task.table_name)]['Keys'].append(task.keys)
            w += 1
            if w >= limit:
                r = dict(query)
                w = 0
                query.clear()
                yield HatsudenkiClient.batch_get_item(r)
        if w is not 0:
            yield HatsudenkiClient.batch_get_item(dict(query))

    async def exec(self, limit=100):
        # TODO: unprocessへの対応がまだ
        ret = defaultdict(list)

        [[ret[key].extend(v) for key, v in r.items()] for r in
         await asyncio.gather(*[q for q in self.exec_query(limit)])]

        return BatchGetResponse(ret)

    @property
    def task_num(self):
        return len(self._task)


class QueryBatchWriteItem(object):
    """
    Batch処理（書き込み）
    """

    def __init__(self):
        self._task: List[BatchWriteTask] = []

    def append(self, task: BatchWriteTask):
        """
        タスク追加

        :param task: 追加するタスクインスタンス
        :return:
        """
        self._task.append(task)

    def append_create(self, table: SoloHatsudenkiTable):
        """
        CREATEのタスクを追加

        :param table: 追加するタスクインスタンス
        :return:
        """
        self._task.append(
            BatchWriteTask(table_name=table._collection_name, item=table.serialize(), kind=BatchWriteKind.Put))

    def append_delete(self, table: SoloHatsudenkiTable):
        """
        DELETEのタスクを追加

        :param table: 追加するタスクインスタンス
        :return:
        """
        t = BatchWriteTask(table_name=table.get_collection_name(), item=table.serialized_key,
                           kind=BatchWriteKind.Delete)
        self._task.append(t)

    def exec_query(self, limit=25):
        task_len = len(self._task)
        chunk_heads = (r * limit for r in range(task_len // limit + (0 < (task_len % limit))))

        def group(head):
            return itertools.groupby(self._task[head:(head + limit)], lambda x: x.real_table_name)

        def query(head):
            return {k: [val.query for val in v] for k, v in group(head)}

        return (HatsudenkiClient.batch_write_item(query(h)) for h in chunk_heads)

    async def exec(self, limit=25):
        g = [q for q in self.exec_query(limit)]
        await asyncio.gather(*g)

    @property
    def task_num(self):
        return len(self._task)


T = TypeVar('T')


class BatchGetResponse(object):
    def __init__(self, result: Dict[str, List[any]]):
        self.result = result

    def __repr__(self):
        return pprint.pformat(self.result)

    def get(self, target_table_type: Type[T]) -> List[T]:
        col_items = self.result[HatsudenkiClient.resolve_table_name(target_table_type.get_collection_name())]
        tt = target_table_type.get_table_type()

        if tt in {TableType.RootTable, TableType.SingleSoloTable, TableType.SingleMultiTable}:
            yield from (target_table_type.deserialize(c) for c in col_items)
            return

        for item in col_items:
            if tt is TableType.ChildTable:
                if item[target_table_type.get_range_key_name()]['S'].startswith(target_table_type.get_tag_name()):
                    yield target_table_type.deserialize(item)
            elif tt is TableType.ChildSoloTable:
                if item[target_table_type.get_range_key_name()]['S'] == target_table_type.get_tag_name():
                    yield target_table_type.deserialize(item)



from collections.__init__ import defaultdict
from enum import Enum
from typing import DefaultDict, List

from hatsudenki.packages.expression.base import BaseExpression


class UpdateReturnType(Enum):
    # 何も返さない
    Nothing = 'NONE'
    # 古い（更新前の）アイテムを返す
    AllOld = 'ALL_OLD'
    # 影響を受けた属性の更新前の値を返す
    UpdatedOld = 'UPDATED_OLD'
    # 更新後のアイテムを返す
    AllNew = 'ALL_NEW'
    # 影響を受けた属性の更新後の値を返す
    UpdatedNew = 'UPDATED_NEW'


class UpdateOperation(Enum):
    # 値のセット
    Set = 'SET'
    # 属性の削除
    Remove = 'REMOVE'
    # 値の加減算（数値型及びセット型のみ）
    Add = 'ADD'
    # セットから値を削除（セット型のみ）
    Delete = 'DELETE'


class UpdateExpression(BaseExpression):
    """
    更新式クラス
    """

    ParameterLabel = 'UpdateExpression'

    def __init__(self):
        super().__init__()
        self.operations: DefaultDict[UpdateOperation, List[str]] = defaultdict(list)

    def set(self, key: str, val: any, raw=False):
        """
        SETオペレーションを追加
        :param key: 対象キー名
        :param val: 対象値
        :param raw: シリアライズされた値が渡されているか
        :return: None
        """
        v = self._register_value(val, raw)
        if v is None:
            return
        k = self._register_key(key)
        exp = f'{k} = {v}'

        self.operations[UpdateOperation.Set].append(exp)

    def increment(self, key: str, val: any, raw=False):
        """
        インクリメント
        内部で'SET #key = #key + :val'を発行するので実質SET
        :param key: 対象キー名
        :param val:
        :param raw:
        :return:
        """
        v = self._register_value(val, raw)
        if v is None:
            return
        k = self._register_key(key)
        exp = f'{k} = {k} + {v}'

        self.operations[UpdateOperation.Set].append(exp)

    def remove(self, *keys: str):
        """
        Removeオペレーションを追加
        :param key: 対象キー名
        :param val: 対象値
        :param raw: シリアライズされた値が渡されているか
        :return: None
        """
        ks = ', '.join(map(self._register_key, keys))
        self.operations[UpdateOperation.Remove].append(ks)

    def remove_raw_key(self, *keys: str):
        ks = ', '.join(keys)
        self.operations[UpdateOperation.Remove].append(ks)

    def add(self, key: str, val: any, raw=False):
        """
        ADDオペレーションを追加
        ADDは数値型及びセットデータ型しかサポートされません
        :param key: 対象キー名
        :param val: 対象値
        :param raw: シリアライズされた値が渡されているか
        :return: None
        """
        v = self._register_value(val, raw)
        if v is None:
            return
        k = self._register_key(key)
        exp = f'{k} {v}'
        self.operations[UpdateOperation.Add].append(exp)

    def delete(self, key: str, val: any, raw=False):
        """
        Deleteオペレーションを追加
        Deleteはデータセット型しかサポートされません
        :param key: 対象キー名
        :param val: 対象値
        :param raw: シリアライズされた値が渡されているか
        :return: None
        """
        if val is None:
            return

        k = self._register_key(key)
        v = self._register_value(val, raw)
        exp = f'{k} {v}'
        self.operations[UpdateOperation.Delete].append(exp)

    def list_append(self, key: str, val: any, raw=False):
        if val is None:
            return

        k = self._register_key(key)
        v = self._register_value(val, raw)

        exp = f'{k} = list_append({k}, {v})'
        self.operations[UpdateOperation.Set].append(exp)

    def list_concat(self, key: str, val: List[any], raw=False):
        if val is None:
            return

        k = self._register_key(key)
        v = self._register_value(val, raw)

        exp = f'list_append({k}, {v})'
        self.operations[UpdateOperation.Set].append(exp)

    @property
    def expression(self):
        """
        AWSにわたす形の表現文字列を取得
        :return: AWSにわたすことのできる表現式文字列
        """
        r = []
        for key, val in self.operations.items():
            r.append(f'{key.value} ' + ', '.join(val))
        return ' '.join(r)

    def dump(self):
        p = self.expression
        for k, v in self.names.items():
            p = p.replace(k, str(v))
        p = p.replace(',', '\n')

        for k, v in self.values.items():
            p = p.replace(k, str(v))

        print(p)



from enum import Enum
from typing import List, Union

from hatsudenki.define.config import KEYS_SEPARATOR
from hatsudenki.packages.field.base import BaseHatsudenkiField


class MasterKeysField(BaseHatsudenkiField['MasterKeysField.Value']):
    TypeStr = 'S'

    class Value(object):
        def __init__(self, name: str, label_list: List[str], parent=None):
            from hatsudenki.packages.table.solo import SoloHatsudenkiTable

            self.parent: SoloHatsudenkiTable = parent
            self.label_list = label_list
            self.name = name

        def is_empty(self):
            return not all(self.label_list)

        def to_string(self):
            return KEYS_SEPARATOR.join(self.label_list)

        def update_key(self):
            if self.parent is None:
                raise Exception('invalid operation. update_key')
            setattr(self.parent, self.name, self)

        def clear(self):
            self.label_list = [
                None
                for _ in range(len(self.label_list))
            ]
            self.update_key()

        def __str__(self):
            return self.to_string()

    def __init__(self, split_num: int, **kwargs):
        super().__init__(**kwargs)
        self.split_num = split_num

    def get_data(self, val: Union[str, list, None, Value], parent=None) -> Value:
        if type(val) is str:
            val = val.split(KEYS_SEPARATOR, self.split_num)
        elif type(val) is list:
            val = self._list_to_cursor(val)
        elif val is None:
            val = [None] * self.split_num
        elif type(val) is self.__class__.Value:
            return val
        else:
            raise Exception(f'invalid keys. type={type(val)}')
        return self.__class__.Value(self.name, val, parent)

    def _list_to_cursor(self, data_list: list):
        # クエリに設定する際に数が足りない要求が来る場合があるので
        if len(data_list) > self.split_num:
            raise Exception(f'data_list size unmatch. recv={len(data_list)} self={self.split_num}')
        ret = []
        for d in data_list:
            if hasattr(d, 'one_cursor'):
                ret.append(d.one_cursor)
            elif hasattr(d, 'value'):
                ret.append(str(d.value))
            else:
                raise Exception('invalid list.')
        return ret

    @classmethod
    def is_empty(cls, value):
        if value is None:
            return True
        if type(value) is list:
            return len(value) is 0
        return value.is_empty()

    def to_string(self, val: Union[str, list, Value]):
        if val is None:
            return None
        if isinstance(val, MasterKeysField.Value):
            return val.to_string()
        if type(val) is str:
            return self.get_data(val).to_string()
        if type(val) is list:
            l = len(val)
            if l is 0:
                return None
            if l > self.split_num:
                raise Exception(f'invalid keys. {val}')

            k = []
            for idx, m in enumerate(val):
                if isinstance(m, Enum):
                    k.append(str(m.value))
                else:
                    k.append(m.one_cursor)

            r = KEYS_SEPARATOR.join(k)
            if l < self.split_num:
                return r + KEYS_SEPARATOR
            else:
                return r
        raise Exception(f'invalid key type {val}')

    def serialize(self, value, table=None):
        if self.__class__.is_empty(value):
            return None
        rv = self.get_data(value, table)
        return {'S': rv.to_string()}

    def deserialize(self, value, table=None):
        s = super().deserialize(value, table)
        return self.get_data(s, table)



class SerializableField(object):
    """
    DynamoDB形式にシリアライズ可能なフィールド
    """

    def serialize(self, data) -> dict:
        """
        DynamoDB形式にシリアライズ

        :param data: any
        :return: dict
        """
        pass

    def deserialize(self, data: dict):
        """
        Python形式にデシリアライズ

        :param data: DynamoDB形式のdict
        :return: any
        """
        pass

    @classmethod
    def is_empty(cls, value):
        return value is None


class Markable(object):
    """
    更新されたアトリビュートを保持可能なオブジェクト
    """

    def modify_mark(self, key):
        """
        更新されたことをマークする

        :param key: 更新されたキー
        :return: None
        """
        pass

    def flush(self):
        """
        変更情報をクリア

        :return: None
        """
        pass



from hatsudenki.packages.command.master.table import MasterTable
from hatsudenki.packages.command.renderer.base import RenderUnit, FileRenderer
from hatsudenki.packages.command.stdout.output import IndentString


class MasterCSModelRenderUnit(RenderUnit[MasterTable]):
    """
    CSのMasterモデルファイルをレンダリングするやつ
    """

    @property
    def record_type(self):
        return f'{self.data.class_name}Record'

    @property
    def list_type(self):
        return f'List<{self.record_type}>'

    def _singleton(self):
        u = IndentString()
        # シングルトン実装
        u.add(f'private static {self.data.class_name} _instance;')
        u.add(f'public static {self.data.class_name} Instance')
        u.indent('{')
        u.add('get')
        u.indent('{')
        u.add('if (_instance == null)')
        u.indent('{')
        u.add(f'_instance = new {self.data.class_name}();')
        u.outdent('}')
        u.add('return _instance;')
        u.outdent('}')
        u.outdent('}')
        return u

    def _records(self):
        u = IndentString()
        hk = self.data.hash_key
        rk = self.data.range_key

        # レコード本体
        if rk:
            t = f'Dictionary<{hk.cs_type_name}, List<{self.record_type}>>'
        else:
            t = f'Dictionary<{hk.cs_type_name}, {self.record_type}>'
        u.add(f'private {t} _records = new {t}();')
        u.add(f'public {t} Records')
        u.indent('{')
        u.add('get')
        u.indent('{')
        u.add('return _records;')
        u.outdent('}')
        u.outdent('}')

        return u

    def _get(self):
        u = IndentString()
        hk = self.data.hash_key
        rk = self.data.range_key

        if rk:
            n = f'{hk.cs_type_name} {hk.column_name}, {rk.cs_type_name} {rk.column_name}'
            c = f'return _records[{hk.column_name}].Find(x => x.GetRangeKey() == {rk.column_name});'
        else:
            n = f'{hk.cs_type_name} {hk.column_name}'
            c = f'return _records[{hk.column_name}];'

        u.add(f'public {self.record_type} Get({n})')
        u.indent('{')
        u.add(f'{c}')
        u.outdent('}')
        return u

    def _get_list(self):
        hk = self.data.hash_key
        rk = self.data.range_key
        if rk is None:
            return None
        u = IndentString()
        n = f'{hk.cs_type_name} {hk.column_name}'
        u.add(f'public List<{self.record_type}> GetList({n})')
        u.indent('{')
        u.add(f'return _records[{hk.column_name}];')
        u.outdent('}')
        return u

    def _find(self, is_all: bool):
        if self.data.range_key is None:
            return None
        u = IndentString()
        hk = self.data.hash_key
        if is_all:
            t = f'List<{self.record_type}>'
            n = 'FindAll'
        else:
            t = self.record_type
            n = 'Find'
        u.add(f'public {t} {n}({hk.cs_type_name} {hk.column_name}, Predicate<{self.record_type}> func)')
        u.indent('{')
        u.add(f'return _records[{hk.column_name}].{n}(func);')
        u.outdent('}')
        return u

    def _filter(self, is_all: bool):
        u = IndentString()
        hk = self.data.hash_key
        if is_all:
            t = f'List<{self.record_type}>'
            n = 'FilterAll'
            fn = 'FindAll'
        else:
            t = self.record_type
            n = 'Filter'
            fn = 'Find'
        u.add(f'public {t} {n}(Predicate<{self.record_type}> func)')
        u.indent('{')
        if is_all:
            u.add(f'List<{self.record_type}> ret = new List<{self.record_type}>();')

        if self.data.range_key is None:
            u.add(f'foreach(KeyValuePair<{hk.cs_type_name}, {self.record_type}> l in _records)')
        else:
            u.add(f'foreach(KeyValuePair<{hk.cs_type_name}, List<{self.record_type}>> l in _records)')
        u.indent('{')

        if is_all:
            if self.data.range_key is None:
                u.add(f'{self.record_type} v = l.Value;')
                u.add('if(func(v))')
                u.indent('{')
                u.add('ret.Add(v);')
                u.outdent('}')
                u.outdent('}')
                u.add('return ret;')
            else:
                u.add(f'List<{self.record_type}> v = l.Value;')
                u.add(f'List<{self.record_type}> t = v.FindAll(func);')
                u.add('if(t.Count > 0)')
                u.indent('{')
                u.add('ret.AddRange(t);')
                u.outdent('}')
                u.outdent('}')
                u.add('return ret;')
        else:
            if self.data.range_key is None:
                u.add(f'{self.record_type} v = l.Value;')
                u.add('if(func(v))')
                u.indent('{')
                u.add('return v;')
                u.outdent('}')
                u.outdent('}')
                u.add('return null;')
            else:
                u.add(f'List<{self.record_type}> v = l.Value;')
                u.add(f'{self.record_type} t = v.Find(func);')
                u.add('if( t != null )')
                u.indent('{')
                u.add('return t;')
                u.outdent('}')
                u.outdent('}')
                u.add('return null;')

        u.outdent('}')

        return u

    def _load(self):
        hk = self.data.hash_key
        u = IndentString()
        u.add(f'public void FinishedLoad({self.record_type}s records)')
        u.indent('{')
        u.add('_records.Clear();')
        u.add(f'foreach({self.record_type} rec in records.Array)')
        u.indent('{')

        u.add(f'{hk.cs_type_name} hk = rec.GetHashKey();')

        if self.data.range_key is None:
            u.add('_records.Add(hk, rec);')
        else:
            u.add('if(!_records.ContainsKey(hk))')
            u.indent('{')
            u.add(f'_records.Add(hk, new List<{self.record_type}>());')
            u.outdent('}')
            u.add('_records[hk].Add(rec);')

        u.outdent('}')
        u.outdent('}')
        return u

    def _repository(self):
        u = IndentString()
        u.add(f'public class {self.data.class_name}')
        u.indent('{')
        u.add(f'public const string TableName = "{self.data.table_name}";')
        # シングルトン実装
        u.add(self._singleton())
        # コンストラクタ
        u.add(f'private {self.data.class_name}() {{ }}')
        # レコード本体
        u.add(self._records())

        # 一件取得
        u.add(self._get())
        u.add(self._get_list())

        # find
        u.add(self._find(is_all=False))
        u.add(self._find(is_all=True))
        # filter
        u.add(self._filter(is_all=False))
        u.add(self._filter(is_all=True))

        # loader
        u.add(self._load())

        u.add(f'public void Release()')
        u.indent('{')
        u.add('_records.Clear();')
        u.add('_records = null;')
        u.add('_instance = null;')
        u.outdent('}')

        u.outdent('}')

        return u

    def _store(self):
        u = IndentString()
        u.add('[MessagePackObject]')
        u.add(f'public class {self.record_type}s')
        u.indent('{')
        u.add('[Key(0)]')
        u.add(f'public {self.record_type}[] Array;')
        u.outdent('}')
        return u

    def _model(self):
        u = IndentString()
        u.add('[MessagePackObject]')
        u.add(f'public class {self.record_type}')
        u.indent('{')

        keys = [c for c in self.data.columns.values() if not c.is_no_pack]
        keys.extend([c for c in self.data.shadow_columns.values() if not c.is_no_pack])

        args = ', '.join((c.cs_argument_name for c in keys))

        u.add(f'public {self.record_type}({args})')
        u.indent('{')
        [u.add(c.cs_assign_name) for c in keys]
        u.outdent('}')

        for idx, c in enumerate(keys):
            if c.is_no_pack:
                continue
            u.add(f'[Key({idx})]', c.cs_define_name)

        hk = self.data.hash_key
        u.add(f'public {hk.cs_type_name} GetHashKey()')
        u.indent('{')
        u.add(f'return {hk.cs_name};')
        u.outdent('}')

        rk = self.data.range_key
        if rk:
            u.add(f'public {rk.cs_type_name} GetRangeKey()')
            u.indent('{')
            u.add(f'return {rk.cs_name};')
            u.outdent('}')

        for c in keys:
            if not c.is_no_pack:
                u.add(c.cs_access_name())

        u.outdent('}')

        return u

    def render(self):
        r = self._repository()
        r.add(self._store())
        r.add(self._model())

        return r


class MasterCSModelRenderer(FileRenderer):

    def render_header(self):
        u = IndentString()
        u.add('using System;')
        u.add('using System.Collections.Generic;')
        u.add('using MessagePack;')
        return u



import unicodedata
from pathlib import Path
from typing import TypeVar, Generic, Dict, Generator, Tuple

from hatsudenki.packages.command.stdout.output import ToolOutput

T = TypeVar('T')


class LoaderBase(Generic[T]):
    def __init__(self, base_path: Path):
        self.base_path = base_path.absolute()
        self.datas: Dict[Path, T] = {}
        self.ext = ''

    def setup(self, dir_name=''):
        ToolOutput.out(f'[{self.__class__.__name__}] setup...')
        self.datas = {Path(unicodedata.normalize('NFC', str(path))): None for path in
                      (self.base_path / dir_name).glob('**/*' + self.ext)}

    def iter(self) -> Generator[Tuple[Path, T], None, None]:
        for k, v in self.datas.items():
            yield k, self.get_from_path(k)

    def get_item(self, file: str) -> T:
        p = self.to_full_path(file)
        return self.get_from_path(p)

    def get_from_path(self, path: Path):
        p = Path(unicodedata.normalize('NFC', str(path.absolute())))
        d = self.datas.get(p)
        if d is None:
            d = self._load(p)
            self.datas[p] = d
        return d

    def exists(self, path: str):
        p = self.to_full_path(path)
        return p in self.datas

    def _load(self, path: Path) -> T:
        raise NotImplementedError()

    def to_full_path(self, path: str):
        p = Path(unicodedata.normalize('NFC', str(self.base_path / path)))
        return p
